import datetime
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import datetimes as dts
from ..api.contexts import MarshalContext
from ..api.contexts import UnmarshalContext
from ..api.types import Marshaler
from ..api.types import Unmarshaler
from ..api.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


DatetimeLikeT = ta.TypeVar('DatetimeLikeT', bound=datetime.datetime | datetime.date | datetime.time)

_DATETIME_LIKES = (
    datetime.datetime,
    datetime.date,
    datetime.time,
)


##


DATE_FORMATS: ta.Sequence[str] = [
    '%Y-%m-%d',
]

TIME_FORMATS: ta.Sequence[str] = [
    ' '.join([tp, *tz])
    for tp in [
        '%H:%M:%S.%f',
        '%H:%M:%S',
        '%H:%M',
    ]
    for tz in ta.cast(list[list[str]], [
        [],
        ['%z'],
        ['%Z'],
        ['%z', '%Z'],
        ['%Z', '%z'],
    ])
]

SEPS: ta.Sequence[str] = ['T', ' ']

DATETIME_FORMATS: ta.Sequence[str] = [
    s.join([df, tf])
    for s in SEPS
    for df in DATE_FORMATS
    for tf in TIME_FORMATS
]


##


@dc.dataclass(frozen=True)
class DatetimeMarshaler(Marshaler, ta.Generic[DatetimeLikeT]):
    cls: type[DatetimeLikeT]
    fmt: str

    def marshal(self, ctx: MarshalContext, o: DatetimeLikeT) -> Value:
        return o.strftime(self.fmt)


_ZERO_DATE = datetime.datetime.now().strptime('', '').date()  # noqa
_ZERO_TIME = datetime.time(0)


@dc.dataclass(frozen=True)
class DatetimeUnmarshaler(Unmarshaler, ta.Generic[DatetimeLikeT]):
    cls: type[DatetimeLikeT]
    fmts: ta.Sequence[str]
    try_iso: bool = False

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> DatetimeLikeT:
        v = check.isinstance(v, str)

        if self.try_iso:
            try:
                return self.cls.fromisoformat(v)  # type: ignore
            except ValueError:
                pass

        for fmt in self.fmts:
            try:
                dt = datetime.datetime.strptime(v, fmt)  # FIXME: timezone  # noqa
            except ValueError:
                pass
            else:
                if self.cls is datetime.datetime:
                    return dt  # type: ignore
                elif self.cls is datetime.date:
                    if dt.time() != _ZERO_TIME:
                        raise ValueError(dt)
                    return dt.date()  # type: ignore
                elif self.cls is datetime.time:
                    if dt.date() != _ZERO_DATE:
                        raise ValueError(dt)
                    return dt.time()  # type: ignore
                else:
                    raise TypeError(self.cls)

        raise ValueError(v)


##


class IsoDatetimeMarshalerUnmarshaler(Marshaler, Unmarshaler, ta.Generic[DatetimeLikeT]):
    cls: type[DatetimeLikeT]

    def marshal(self, ctx: MarshalContext, o: DatetimeLikeT) -> Value:
        return o.isoformat()

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> DatetimeLikeT:
        return self.cls.fromisoformat(v)  # type: ignore


##


class TimedeltaMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: datetime.timedelta) -> Value:
        return str(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return dts.parse_timedelta(check.isinstance(v, str))


##


DATETIME_MARSHALER_FACTORY = TypeMapMarshalerFactory({
    **{cls: DatetimeMarshaler(cls, DATETIME_FORMATS[0]) for cls in _DATETIME_LIKES},
    datetime.timedelta: TimedeltaMarshalerUnmarshaler(),
})

DATETIME_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({
    **{cls: DatetimeUnmarshaler(cls, DATETIME_FORMATS, try_iso=True) for cls in _DATETIME_LIKES},
    datetime.timedelta: TimedeltaMarshalerUnmarshaler(),
})



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
from ....lite.check import check
from ..compression.compressors import IoPipelineHttpObjectCompressor
from ..encoders import IoPipelineHttpObjectEncoder
from ..objects import IoPipelineHttpMessageHead
from ..requests import IoPipelineHttpRequestHead
from ..requests import IoPipelineHttpRequestObjects


##


class IoPipelineHttpRequestEncoder(IoPipelineHttpRequestObjects, IoPipelineHttpObjectEncoder):
    def _encode_head_line(self, head: IoPipelineHttpMessageHead) -> bytes:
        head = check.isinstance(head, IoPipelineHttpRequestHead)
        version_str = f'HTTP/{head.version.major}.{head.version.minor}'
        return f'{head.method} {head.target} {version_str}\r\n'.encode('ascii')


##


class IoPipelineHttpRequestCompressor(IoPipelineHttpRequestObjects, IoPipelineHttpObjectCompressor):
    pass



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import functools
import typing as ta

from ....lite.abstract import Abstract
from ....lite.check import check
from ....lite.namespaces import NamespaceClass
from ..core import IoPipelineDirectionOrDuplex
from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineHandlerFn
from ..core import IoPipelineMessages
from .fns import IoPipelineHandlerFns


##


FlatMapIoPipelineHandlerFn = IoPipelineHandlerFn[ta.Any, ta.Iterable[ta.Any]]  # ta.TypeAlias  # omlish-amalg-typing-no-move  # noqa


class FlatMapIoPipelineHandlerFns(NamespaceClass):
    @dc.dataclass(frozen=True)
    class Filter:
        pred: IoPipelineHandlerFn[ta.Any, bool]
        fn: FlatMapIoPipelineHandlerFn
        else_fn: ta.Optional[FlatMapIoPipelineHandlerFn] = None

        def __repr__(self) -> str:
            return (
                f'{type(self).__name__}('
                f'{self.pred!r}'
                f', {self.fn!r}'
                f'{f", else_fn={self.else_fn!r}" if self.else_fn is not None else ""}'
                f')'
            )

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            if self.pred(ctx, msg):
                yield from self.fn(ctx, msg)
            elif (ef := self.else_fn) is not None:
                yield from ef(ctx, msg)  # noqa
            else:
                yield msg

    @classmethod
    def filter(
            cls,
            pred: IoPipelineHandlerFn[ta.Any, bool],
            fn: FlatMapIoPipelineHandlerFn,
            else_fn: ta.Optional[FlatMapIoPipelineHandlerFn] = None,
    ) -> FlatMapIoPipelineHandlerFn:
        return cls.Filter(pred, fn, else_fn)

    @classmethod
    def filter_type(
            cls,
            ty: ta.Union[type, ta.Tuple[type, ...]],
            fn: FlatMapIoPipelineHandlerFn,
            else_fn: ta.Optional[FlatMapIoPipelineHandlerFn] = None,
    ) -> FlatMapIoPipelineHandlerFn:
        return cls.filter(IoPipelineHandlerFns.isinstance(ty), fn, else_fn)

    #

    @dc.dataclass(frozen=True)
    class Concat:
        fns: ta.Sequence[FlatMapIoPipelineHandlerFn]

        def __repr__(self) -> str:
            return f'{type(self).__name__}([{", ".join(map(repr, self.fns))}])'

        def __post_init__(self) -> None:
            check.not_empty(self.fns)

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            for fn in self.fns:
                yield from fn(ctx, msg)

    @classmethod
    def concat(cls, *fns: FlatMapIoPipelineHandlerFn) -> FlatMapIoPipelineHandlerFn:
        return cls.Concat(fns)

    #

    @dc.dataclass(frozen=True)
    class Compose:
        fns: ta.Sequence[FlatMapIoPipelineHandlerFn]

        def __repr__(self) -> str:
            return f'{type(self).__name__}([{", ".join(map(repr, self.fns))}])'

        _fn: FlatMapIoPipelineHandlerFn = dc.field(init=False)

        def __post_init__(self) -> None:
            check.not_empty(self.fns)

            def compose(cur, nxt, ctx, msg):
                for x in cur(ctx, msg):
                    yield from nxt(ctx, x)

            xf: ta.Any = lambda ctx, msg: (msg,)  # noqa
            for cf in reversed(self.fns):
                xf = functools.partial(compose, cf, xf)

            object.__setattr__(self, '_fn', xf)

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            return self._fn(ctx, msg)

    @classmethod
    def compose(cls, *fns: FlatMapIoPipelineHandlerFn) -> FlatMapIoPipelineHandlerFn:
        return cls.Compose(fns)

    #

    @dc.dataclass(frozen=True)
    class Map:
        fn: IoPipelineHandlerFn[ta.Any, ta.Any]

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.fn!r})'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            return (self.fn(ctx, msg),)

    @classmethod
    def map(cls, fn: IoPipelineHandlerFn[ta.Any, ta.Any]) -> FlatMapIoPipelineHandlerFn:
        return cls.Map(fn)

    #

    @dc.dataclass(frozen=True)
    class Apply:
        fn: IoPipelineHandlerFn[ta.Any, None]

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.fn!r})'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            self.fn(ctx, msg)
            return (msg,)

    @classmethod
    def apply(cls, fn: IoPipelineHandlerFn[ta.Any, None]) -> FlatMapIoPipelineHandlerFn:
        return cls.Apply(fn)

    #

    @dc.dataclass(frozen=True)
    class Feed:
        direction: IoPipelineDirectionOrDuplex

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.direction!r})'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            if self.direction == 'inbound':
                ctx.feed_in(msg)
            elif self.direction == 'outbound':
                ctx.feed_out(msg)
            else:
                raise RuntimeError(f'Unknown direction: {self.direction!r}')
            return (msg,)

    @classmethod
    def feed_in(cls) -> FlatMapIoPipelineHandlerFn:
        return cls.Feed('inbound')

    @classmethod
    def feed_out(cls) -> FlatMapIoPipelineHandlerFn:
        return cls.Feed('outbound')

    #

    @dc.dataclass(frozen=True)
    class Inject:
        before: ta.Union[ta.Sequence[ta.Any], ta.Callable[[], ta.Sequence[ta.Any]], None] = None
        after: ta.Union[ta.Sequence[ta.Any], ta.Callable[[], ta.Sequence[ta.Any]], None] = None

        def __repr__(self) -> str:
            return ''.join([
                f'{type(self).__name__}(',
                *', '.join([
                    *([f'before={self.before!r}'] if self.before is not None else []),
                    *([f'after={self.after!r}'] if self.after is not None else []),
                ]),
                ')',
            ])

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            out: ta.List[ta.Any] = []
            if (before := self.before) is not None:
                out.extend(before() if callable(before) else before)  # noqa
            out.append(msg)
            if (after := self.after) is not None:
                out.extend(after() if callable(after) else after)  # noqa
            return out

    @classmethod
    def inject(
            cls,
            *,
            before: ta.Union[ta.Sequence[ta.Any], ta.Callable[[], ta.Sequence[ta.Any]], None] = None,
            after: ta.Union[ta.Sequence[ta.Any], ta.Callable[[], ta.Sequence[ta.Any]], None] = None,
    ) -> FlatMapIoPipelineHandlerFn:
        return cls.Inject(
            before=before,
            after=after,
        )

    #

    @dc.dataclass(frozen=True)
    class Drop:
        def __repr__(self) -> str:
            return f'{type(self).__name__}()'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            return ()

    @classmethod
    def drop(cls) -> FlatMapIoPipelineHandlerFn:
        return cls.Drop()

    #

    @dc.dataclass(frozen=True)
    class Nop:
        def __repr__(self) -> str:
            return f'{type(self).__name__}()'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> ta.Iterable[ta.Any]:
            return (msg,)

    @classmethod
    def nop(cls) -> FlatMapIoPipelineHandlerFn:
        return cls.Nop()


#


class FlatMapIoPipelineHandler(IoPipelineHandler, Abstract):
    def __init__(
            self,
            fn: FlatMapIoPipelineHandlerFn,
    ) -> None:
        super().__init__()

        self._fn = check.callable(fn)

    _fn: FlatMapIoPipelineHandlerFn

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}({self._fn!r})'


#


class InboundFlatMapIoPipelineHandler(FlatMapIoPipelineHandler):
    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        for x in self._fn(ctx, msg):
            ctx.feed_in(x)


class OutboundFlatMapIoPipelineHandler(FlatMapIoPipelineHandler):
    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        for x in self._fn(ctx, msg):
            ctx.feed_out(x)


class DuplexFlatMapIoPipelineHandler(
    InboundFlatMapIoPipelineHandler,
    OutboundFlatMapIoPipelineHandler,
):
    pass


#


class FlatMapIoPipelineHandlers(NamespaceClass):
    _CLS_BY_DIRECTION: ta.ClassVar[ta.Mapping[IoPipelineDirectionOrDuplex, ta.Type[FlatMapIoPipelineHandler]]] = {  # noqa
        'inbound': InboundFlatMapIoPipelineHandler,
        'outbound': OutboundFlatMapIoPipelineHandler,
        'duplex': DuplexFlatMapIoPipelineHandler,
    }

    @classmethod
    def new(
            cls,
            direction: IoPipelineDirectionOrDuplex,
            fn: FlatMapIoPipelineHandlerFn,
    ) -> IoPipelineHandler:
        h_cls = cls._CLS_BY_DIRECTION[direction]
        return h_cls(fn)

    #

    _NOT_MUST_PROPAGATE: ta.ClassVar[IoPipelineHandlerFn[ta.Any, bool]] = IoPipelineHandlerFns.not_(
        IoPipelineHandlerFns.isinstance(IoPipelineMessages.MustPropagate),
    )

    @classmethod
    def add_filters(
            cls,
            fn: FlatMapIoPipelineHandlerFn,
            *,
            filter_type: ta.Optional[ta.Union[type, ta.Tuple[type, ...]]] = None,
            filter: ta.Optional[IoPipelineHandlerFn[ta.Any, bool]] = None,  # noqa
            not_must_propagate: bool = False,
    ) -> FlatMapIoPipelineHandlerFn:
        if filter is not None:
            fn = FlatMapIoPipelineHandlerFns.filter(filter, fn)

        if filter_type is not None:
            fn = FlatMapIoPipelineHandlerFns.filter_type(filter_type, fn)

        if not_must_propagate:
            fn = FlatMapIoPipelineHandlerFns.filter(cls._NOT_MUST_PROPAGATE, fn)

        return fn

    #

    @classmethod
    def drop(
            cls,
            direction: IoPipelineDirectionOrDuplex,
            *,
            filter_type: ta.Optional[ta.Union[type, ta.Tuple[type, ...]]] = None,
            filter: ta.Optional[IoPipelineHandlerFn[ta.Any, bool]] = None,  # noqa
    ) -> IoPipelineHandler:
        return cls.new(
            direction,
            cls.add_filters(
                FlatMapIoPipelineHandlerFns.drop(),
                filter=filter,
                filter_type=filter_type,
                not_must_propagate=True,
            ),
        )

    @classmethod
    def apply_and_drop(
            cls,
            direction: IoPipelineDirectionOrDuplex,
            fn: IoPipelineHandlerFn[ta.Any, None],
            *,
            filter_type: ta.Optional[ta.Union[type, ta.Tuple[type, ...]]] = None,
            filter: ta.Optional[IoPipelineHandlerFn[ta.Any, bool]] = None,  # noqa
    ) -> IoPipelineHandler:
        return cls.new(
            direction,
            cls.add_filters(
                FlatMapIoPipelineHandlerFns.compose(
                    FlatMapIoPipelineHandlerFns.apply(fn),
                    FlatMapIoPipelineHandlerFns.drop(),
                ),
                filter=filter,
                filter_type=filter_type,
                not_must_propagate=True,
            ),
        )

    @classmethod
    def feed_out_and_drop(
            cls,
            *,
            filter_type: ta.Optional[ta.Union[type, ta.Tuple[type, ...]]] = None,
            filter: ta.Optional[IoPipelineHandlerFn[ta.Any, bool]] = None,  # noqa
    ) -> IoPipelineHandler:
        return cls.new(
            'inbound',
            cls.add_filters(
                FlatMapIoPipelineHandlerFns.compose(
                    FlatMapIoPipelineHandlerFns.feed_out(),
                    FlatMapIoPipelineHandlerFns.drop(),
                ),
                filter=filter,
                filter_type=filter_type,
                not_must_propagate=True,
            ),
        )



import dataclasses as dc
import typing as ta

from ..... import check
from ..... import lang
from .....lite.dataclasses import is_immediate_dataclass
from ....specs import InitFn
from ....specs import ReprFn


T = ta.TypeVar('T')


##


METADATA_ATTR = '__dataclass_metadata__'


def _get_cls_metadata_dct(cls: type) -> dict:
    check.isinstance(cls, type)
    if is_immediate_dataclass(cls):
        raise TypeError(f'Cannot alter dataclass metadata on already processed class {cls!r}')
    try:
        return cls.__dict__[METADATA_ATTR]
    except KeyError:
        pass
    dct: dict = {}
    setattr(cls, METADATA_ATTR, dct)
    return dct


def _append_cls_metadata(cls, k, v):
    _get_cls_metadata_dct(cls).setdefault(k, []).append(v)


def _append_cls_dct_metadata(k, *vs):
    lang.get_caller_cls_dct(1).setdefault(METADATA_ATTR, {}).setdefault(k, []).extend(vs)


##


class _ExtraClassParamsMetadata(lang.Marker):
    pass


# The following function duplicates non-stdlib kwargs/defaults from the class decorator, but does so only for
# autocomplete - the underlying impl blindly forwards given kwargs to allow distinction between `None` and absent.


def extra_class_params(
        *,

        metadata: ta.Sequence[ta.Any] | None = None,  # noqa

        reorder: bool | None = None,
        cache_hash: bool | None = None,
        generic_init: bool | None = None,
        override: bool | None = None,
        allow_dynamic_dunder_attrs: bool | None = None,

        repr_id: bool | None = None,
        terse_repr: bool | None = None,
        default_repr_fn: ReprFn | None = None,

        allow_redundant_decorator: bool | None = None,
):
    def inner(cls):
        return cls
    return inner


def _extra_class_params(**kwargs):
    def inner(cls):
        _append_cls_metadata(cls, _ExtraClassParamsMetadata, kwargs)
        return cls
    return inner


globals()['extra_class_params'] = _extra_class_params


##


class _UserMetadata(lang.Marker):
    pass


def metadata(*objs) -> None:
    _append_cls_dct_metadata(_InitMetadata, *objs)


def append_class_metadata(cls: type[T], *args: ta.Any) -> type[T]:
    _append_cls_metadata(cls, _UserMetadata, *args)
    return cls


#


class _InitMetadata(lang.Marker):
    pass


def init(obj):
    _append_cls_dct_metadata(_InitMetadata, obj)
    return obj


#


class _ValidateMetadata(lang.Marker):
    pass


def validate(obj):
    _append_cls_dct_metadata(_ValidateMetadata, obj)
    return obj


##


@dc.dataclass(frozen=True, kw_only=True, eq=False)
class ClassMetadata:
    extra_params: ta.Mapping[str, ta.Any] | None = None

    user_metadata: ta.Sequence[ta.Any] | None = None
    init_fns: ta.Sequence[InitFn | property] | None = None
    validate_fns: ta.Sequence[ta.Any] | None = None


def extract_cls_metadata(
        cls: type,
        *,
        deep: bool,
) -> ClassMetadata:
    extra_params = {}

    user_metadata: list[ta.Any] = []
    init_fns: list[InitFn | property] = []
    validate_fns: list[ta.Any] = []

    for b_cls in (cls.__mro__[-2::-1] if deep else (cls,)):
        cls_md_dct = b_cls.__dict__.get(METADATA_ATTR, {})

        if b_cls is cls:
            for kw in cls_md_dct.get(_ExtraClassParamsMetadata, []):
                extra_params.update(kw)

        user_metadata.extend(cls_md_dct.get(_UserMetadata) or [])
        init_fns.extend(cls_md_dct.get(_InitMetadata) or [])
        validate_fns.extend(cls_md_dct.get(_ValidateMetadata) or [])

    return ClassMetadata(
        extra_params=extra_params or None,

        user_metadata=user_metadata,
        init_fns=init_fns,
        validate_fns=validate_fns,
    )



import typing as ta

from ... import inject as inj
from .. import asgi
from .. import sessions
from .base import SCOPE
from .markers import AppMarker
from .markers import AppMarkerProcessor
from .markers import NopAppMarkerProcessor
from .routes import Route
from .routes import RouteHandlerHolder
from .routes import _HandlesAppMarker
from .routes import build_route_handler_map
from .sessions import SESSION
from .sessions import _WithSessionAppMarker
from .sessions import _WithSessionAppMarkerProcessor
from .templates import TemplateNamespace
from .templates import Templates


##


def bind_route_handler_class(hc: type[RouteHandlerHolder]) -> inj.Elemental:
    return inj.as_elements(
        inj.bind(hc, singleton=True),
        inj.set_binder[RouteHandlerHolder]().bind(hc),
    )


def bind_app_marker_processor(mc: type[AppMarker], pc: type[AppMarkerProcessor]) -> inj.Elemental:
    return inj.as_elements(
        inj.bind(pc),
        inj.map_binder[type[AppMarker], AppMarkerProcessor]().bind(mc, pc),
    )


def _build_route_handler_map(
        handler_holders: ta.AbstractSet[RouteHandlerHolder],
        processors: ta.Mapping[type[AppMarker], AppMarkerProcessor],
) -> ta.Mapping[Route, asgi.App]:
    return build_route_handler_map(
        handler_holders,
        processors,
    )


def bind_route_handler_map() -> inj.Elemental:
    return inj.as_elements(
        inj.bind(_build_route_handler_map, singleton=True),
        inj.map_binder[type[AppMarker], AppMarkerProcessor](),
    )


def bind() -> inj.Elemental:
    return inj.as_elements(
        inj.bind(ta.Callable[[], asgi.Scope], to_const=SCOPE.get),
        inj.bind(ta.Callable[[], sessions.Session], to_const=SESSION.get),

        ##

        inj.map_binder[type[AppMarker], AppMarkerProcessor](),

        bind_app_marker_processor(_WithSessionAppMarker, _WithSessionAppMarkerProcessor),
        bind_app_marker_processor(_HandlesAppMarker, NopAppMarkerProcessor),

        ##

        inj.set_binder[RouteHandlerHolder](),
    )


def _build_template_namespaces(
        ns: ta.Annotated[ta.Mapping[str, ta.Any], inj.Tag(TemplateNamespace)],
) -> TemplateNamespace:
    return TemplateNamespace(ns)


def bind_templates(impl: type[Templates]) -> inj.Elemental:
    return inj.as_elements(
        inj.bind(impl, singleton=True),
        inj.bind(Templates, to_key=impl),

        inj.map_binder[str, ta.Any](tag=TemplateNamespace),
        inj.bind(_build_template_namespaces),
    )



# ruff: noqa: UP043 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from .asyncs import AsyncHttpClient
from .asyncs import AsyncStreamHttpClientResponse
from .base import HttpClientContext
from .base import HttpClientRequest
from .sync import HttpClient
from .sync import StreamHttpClientResponse


##


class ExecutorAsyncHttpClient(AsyncHttpClient):
    def __init__(
            self,
            run_in_executor: ta.Callable[..., ta.Awaitable],
            client: HttpClient,
    ) -> None:
        super().__init__()

        self._run_in_executor = run_in_executor
        self._client = client

    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        owner: 'ExecutorAsyncHttpClient'
        resp: StreamHttpClientResponse

        async def read1(self, n: int = -1, /) -> bytes:
            return await self.owner._run_in_executor(self.resp.stream.read1, n)  # noqa

        async def read(self, n: int = -1, /) -> bytes:
            return await self.owner._run_in_executor(self.resp.stream.read, n)  # noqa

        async def close(self) -> None:
            return await self.owner._run_in_executor(self.resp.close)  # noqa

    async def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> AsyncStreamHttpClientResponse:
        resp: StreamHttpClientResponse = await self._run_in_executor(lambda: self._client.stream_request(req, context=ctx))  # noqa
        return AsyncStreamHttpClientResponse(
            status=resp.status,
            headers=resp.headers,
            request=req,
            underlying=resp,
            **(dict(  # type: ignore
                _stream=(adapter := self._StreamAdapter(self, resp)),
                _closer=adapter.close,
            ) if resp.has_data else {}),
        )



# ruff: noqa: UP045
import typing as ta

import anyio

from ..events import AsyncliteEvent
from ..events import AsyncliteEvents
from .base import AnyioAsyncliteApi
from .base import AnyioAsyncliteObject


##


class AnyioAsyncliteEvent(AsyncliteEvent, AnyioAsyncliteObject):
    def __init__(self, u: anyio.Event) -> None:
        super().__init__()

        self._u = u

    def set(self) -> None:
        return self._u.set()

    def is_set(self) -> bool:
        return self._u.is_set()

    async def wait(self, *, timeout: ta.Optional[float] = None) -> None:
        if timeout is not None:
            with anyio.fail_after(timeout):
                await self._u.wait()

        else:
            await self._u.wait()


class AnyioAsyncliteEvents(AsyncliteEvents, AnyioAsyncliteApi):
    def make_event(self) -> AsyncliteEvent:
        return AnyioAsyncliteEvent(anyio.Event())



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import asyncio
import dataclasses as dc
import typing as ta

from ....io.pipelines.core import IoPipelineHandlerContext
from ....io.pipelines.core import IoPipelineMessages
from ....io.pipelines.drivers.asyncio import SimpleAsyncioStreamIoPipelineDriver
from ....io.pipelines.handlers.flatmap import FlatMapIoPipelineHandlers
from ....io.readers import AsyncBytesReader
from ....io.readers import AsyncBytesReaders
from ....lite.check import check
from ...clients.asyncs import AsyncHttpClient
from ...clients.asyncs import AsyncStreamHttpClientResponse
from ...clients.base import HttpClientContext
from ...clients.base import HttpClientRequest
from ...pipelines.clients.clients import IoPipelineHttpClientMessages
from ...pipelines.responses import FullIoPipelineHttpResponse
from .base import BaseIoPipelineHttpClient


##


class AsyncioIoPipelineAsyncHttpClient(AsyncHttpClient, BaseIoPipelineHttpClient):
    class _DriverResponseReader:
        def __init__(
                self,
                drv: SimpleAsyncioStreamIoPipelineDriver,
        ) -> None:
            super().__init__()

            self._drv = drv

            self._done = False

        async def read1(self, n: int = -1, /) -> bytes:
            raise NotImplementedError

        async def read(self, n: int = -1, /) -> bytes:
            raise NotImplementedError

    #

    async def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> AsyncStreamHttpClientResponse:
        prepared = self._prepare_request(req)

        #

        response: ta.Optional[FullIoPipelineHttpResponse] = None

        def on_output(h_ctx: IoPipelineHandlerContext, out: IoPipelineHttpClientMessages.Output) -> None:
            msg = out.msg

            if isinstance(msg, FullIoPipelineHttpResponse):
                nonlocal response
                check.none(response)
                response = msg

                h_ctx.pipeline.feed_in(IoPipelineHttpClientMessages.Close())

            elif isinstance(msg, (IoPipelineMessages.FinalInput, IoPipelineHttpClientMessages.Close)):
                pass

            else:
                raise TypeError(msg)

        pipeline_spec = dc.replace(
            prepared.pipeline_spec,
            handlers=[
                FlatMapIoPipelineHandlers.apply_and_drop(
                    'outbound',
                    on_output,
                    filter_type=IoPipelineHttpClientMessages.Output,
                ),
                *prepared.pipeline_spec.handlers,
            ],
        )

        #

        reader, writer = await asyncio.open_connection(prepared.parsed_url.host, prepared.parsed_url.port)

        try:
            drv = SimpleAsyncioStreamIoPipelineDriver(
                pipeline_spec,
                reader,
                writer,
            )

            drv_run_task = asyncio.create_task(drv.run())

            await drv.enqueue_waitable(IoPipelineHttpClientMessages.Request(
                prepared.full_request,
                # aggregate=...
            ))

            await drv_run_task

            #

            response = check.not_none(response)

            response_reader: AsyncBytesReader

            if isinstance(response, FullIoPipelineHttpResponse):
                response_reader = AsyncBytesReaders.of_bytes(response.body)

                drv.pipeline.destroy()

                writer.close()
                await writer.wait_closed()

                async def close() -> None:
                    pass

            else:
                response_reader = self._DriverResponseReader(drv)  # type: ignore[unreachable]

                async def close() -> None:
                    try:
                        drv.pipeline.destroy()
                    finally:
                        writer.close()
                        await writer.wait_closed()

            #

            head = check.not_none(response).head

            return AsyncStreamHttpClientResponse(
                status=head.status,
                headers=head.headers,
                request=req,
                underlying=drv,
                _stream=response_reader,
                _closer=close,
            )

        except BaseException:
            writer.close()
            await writer.wait_closed()

            raise



import typing as ta

import pytest


T = ta.TypeVar('T')


##


def find_plugin(pm: pytest.PytestPluginManager, ty: type[T]) -> T | None:
    lst = [p for p in pm.get_plugins() if isinstance(p, ty)]
    if not lst:
        return None
    [ret] = lst
    return ret



# ruff: noqa: FURB188 UP006 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ...io.pipelines.core import IoPipelineHandler
from ...io.pipelines.core import IoPipelineHandlerContext
from ...io.pipelines.core import IoPipelineMessages


##


@dc.dataclass(frozen=True)
class IoPipelineSseEvent:
    event: ta.Optional[str] = None
    data: str = ''
    id: ta.Optional[str] = None
    retry: ta.Optional[int] = None


##


class IoPipelineSseDecoder(IoPipelineHandler):
    """Consumes lines and emits SseEvent objects; ignores comment lines and handles blank-line termination."""

    def __init__(self) -> None:
        super().__init__()

        self._event: ta.Optional[str] = None
        self._data: ta.List[str] = []
        self._id: ta.Optional[str] = None
        self._retry: ta.Optional[int] = None

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineMessages.FinalInput):
            self._emit_if_any(ctx)
            ctx.feed_in(msg)
            return

        if not isinstance(msg, str):
            ctx.feed_in(msg)
            return

        line = msg

        if line.endswith('\r'):
            line = msg[:-1]

        if not line:
            self._emit_if_any(ctx)
            return

        if line.startswith(':'):
            return

        if ':' in line:
            field, value = line.split(':', 1)
            if value.startswith(' '):
                value = value[1:]
        else:
            field, value = line, ''

        if field == 'event':
            self._event = value
        elif field == 'data':
            self._data.append(value)
        elif field == 'id':
            self._id = value
        elif field == 'retry':
            try:
                self._retry = int(value)
            except ValueError:
                pass

    def _emit_if_any(self, ctx: IoPipelineHandlerContext) -> None:
        if (
                self._event is None and
                not self._data and
                self._id is None and
                self._retry is None
        ):
            return

        ev = IoPipelineSseEvent(
            event=self._event,
            data='\n'.join(self._data),
            id=self._id,
            retry=self._retry,
        )

        self._event = None
        self._data.clear()
        self._id = None
        self._retry = None

        ctx.feed_in(ev)



import datetime
import re
import typing as ta


##


def _is_exact_int(x: object) -> bool:
    return type(x) is int


def _is_exact_float(x: object) -> bool:
    return type(x) is float


def _is_exact_int_or_float(x: object) -> bool:
    return type(x) in (int, float)


##


@ta.final
class Bytes:
    """
    - +, - with exact int or Bytes -> Bytes
    - * with exact int -> Bytes
    - / and // with exact int -> Bytes (only if divisible exactly)
    - disallows Bytes*Bytes, Bytes/Bytes, etc.
    """

    def __new__(cls, v: ta.Union[int, 'Bytes']) -> 'Bytes':
        if isinstance(v, cls):
            return v
        else:
            return super().__new__(cls)

    def __init__(self, v: ta.Union['Bytes', int]) -> None:
        if isinstance(v, Bytes):
            if v is not self:
                raise TypeError(v)
        else:
            self._v = int(v)

    def __int__(self) -> int:
        return self._v

    # formatting / parsing

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._v!r})'

    _DEC_UNITS: ta.ClassVar[tuple[tuple[str, int], ...]] = (
        ('B', 1),
        ('KB', 1000),
        ('MB', 1000**2),
        ('GB', 1000**3),
        ('TB', 1000**4),
        ('PB', 1000**5),
        ('EB', 1000**6),
    )

    _BIN_UNITS: ta.ClassVar[tuple[tuple[str, int], ...]] = (
        ('B', 1),
        ('KiB', 1024),
        ('MiB', 1024**2),
        ('GiB', 1024**3),
        ('TiB', 1024**4),
        ('PiB', 1024**5),
        ('EiB', 1024**6),
    )

    _UNIT_MAP: ta.ClassVar[ta.Mapping[str,int]] = dict(_DEC_UNITS + _BIN_UNITS)

    def __str__(self) -> str:
        n = self._v
        sign = '-' if n < 0 else ''
        v = abs(n)
        if not v:
            return '0 B'

        # choose binary for nice powers-of-two, else decimal.
        # heuristic: prefer binary if divisible by 1024 and >= 1024.
        units = self._BIN_UNITS if (v >= 1024 and v % 1024 == 0) else self._DEC_UNITS

        # pick the largest unit that divides evenly; otherwise fall back to B.
        chosen_name, chosen_scale = 'B', 1
        for name, scale in units:
            if scale == 1:
                continue
            if v % scale == 0:
                chosen_name, chosen_scale = name, scale

        if chosen_scale == 1:
            return f'{sign}{v} B'
        return f'{sign}{v // chosen_scale} {chosen_name}'

    _PARSE_RE: ta.ClassVar[re.Pattern[str]] = re.compile(
        r'^\s*([+-]?\d+)\s*([A-Za-z]+)\s*$',
    )

    @classmethod
    def parse(cls, s: str) -> 'Bytes':
        m = cls._PARSE_RE.match(s)
        if not m:
            raise ValueError(f'Invalid Bytes string: {s!r}')
        num_s, unit = m.group(1), m.group(2)

        num = int(num_s)

        try:
            scale = cls._UNIT_MAP[unit]
        except KeyError:
            raise ValueError(f'Unknown byte unit {unit!r}') from None

        return cls(num * scale)

    # operators

    def __hash__(self) -> int:
        return hash(self._v)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Bytes):
            return NotImplemented
        return self._v == other._v

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Bytes):
            return NotImplemented
        return self._v != other._v

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Bytes):
            return NotImplemented
        return self._v < other._v

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Bytes):
            return NotImplemented
        return self._v <= other._v

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Bytes):
            return NotImplemented
        return self._v > other._v

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Bytes):
            return NotImplemented
        return self._v >= other._v

    def _coerce_other(self, other: object) -> int:
        if isinstance(other, Bytes):
            return other._v  # Noqa
        elif _is_exact_int(other):
            return other  # type: ignore[return-value]
        else:
            raise TypeError(f'Bytes can only add/sub Bytes or exact int, got {type(other).__name__}')

    def __add__(self, other: object) -> 'Bytes':
        return Bytes(self._v + self._coerce_other(other))

    def __radd__(self, other: object) -> 'Bytes':
        return self.__add__(other)

    def __sub__(self, other: object) -> 'Bytes':
        return Bytes(self._v - self._coerce_other(other))

    def __rsub__(self, other: object) -> 'Bytes':
        if isinstance(other, Bytes):
            return Bytes(other._v - self._v)
        if _is_exact_int(other):
            return Bytes(other - self._v)  # type: ignore[operator]
        return NotImplemented  # will raise TypeError

    def __mul__(self, other: object) -> 'Bytes':
        if _is_exact_int(other):
            return Bytes(self._v * other)  # type: ignore[operator]
        raise TypeError(f'Bytes can only be multiplied by exact int, got {type(other).__name__}')

    def __rmul__(self, other: object) -> 'Bytes':
        return self.__mul__(other)

    def _div_exact(self, other: object) -> 'Bytes':
        if not _is_exact_int(other):
            raise TypeError(f'Bytes can only be divided by exact int, got {type(other).__name__}')
        if other == 0:
            raise ZeroDivisionError
        n = self._v
        q, r = divmod(n, other)  # type: ignore[operator]
        if r != 0:
            raise ValueError(f'Non-exact byte division: {n} / {other} has remainder {r}')
        return Bytes(q)

    def __truediv__(self, other: object) -> 'Bytes':
        return self._div_exact(other)

    def __floordiv__(self, other: object) -> 'Bytes':
        return self._div_exact(other)


B: ta.Final[Bytes] = Bytes(1)

KB: ta.Final[Bytes] = 1000 * B
MB: ta.Final[Bytes] = 1000 * KB
GB: ta.Final[Bytes] = 1000 * MB
TB: ta.Final[Bytes] = 1000 * GB

KiB: ta.Final[Bytes] = 1024 * B
MiB: ta.Final[Bytes] = 1024 * KiB
GiB: ta.Final[Bytes] = 1024 * MiB
TiB: ta.Final[Bytes] = 1024 * GiB


##


@ta.final
class Seconds:
    """
    - +, - with exact int/float or Seconds -> Seconds
    - * and / with exact int/float -> Seconds
    - disallows Seconds*Seconds, Seconds/Seconds, etc.
    - timedelta interop via to_timedelta / from_timedelta
    """

    def __new__(cls, v: ta.Union[float, 'Seconds']) -> 'Seconds':
        if isinstance(v, cls):
            return v
        else:
            return super().__new__(cls)

    def __init__(self, v: ta.Union['Seconds', float]) -> None:
        if isinstance(v, Seconds):
            if v is not self:
                raise TypeError(v)
        else:
            self._v = float(v)

    def __float__(self) -> float:
        return self._v

    # formatting / parsing

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._v!r})'

    _UNITS: ta.ClassVar[tuple[tuple[str, float], ...]] = (
        ('ns', 1e-9),
        ('us', 1e-6),
        ('ms', 1e-3),
        ('s', 1.0),
        ('m', 60.0),
        ('h', 3600.0),
        ('d', 86400.0),
    )

    _UNITS_MAP: ta.ClassVar[ta.Mapping[str, float]] = dict(_UNITS)

    def __str__(self) -> str:
        x = self._v
        if not x:
            return '0 s'

        sign = '-' if x < 0 else ''
        v = abs(x)

        # choose a "nice" unit:
        # - if >= 60, prefer d/h/m where the scaled value is >= 1
        # - if < 1, prefer ms/us/ns where the scaled value is >= 1
        # For sub-second, pick the *largest* sub-second unit that yields >= 1.
        unit_name, unit_scale = 's', 1.0

        if v >= 60.0:
            for name, scale in reversed(self._UNITS):
                if scale >= 1.0 and (v / scale) >= 1.0:
                    unit_name, unit_scale = name, scale
                    break
        elif v < 1.0:
            for name, scale in reversed(self._UNITS):
                if scale < 1.0 and (v / scale) >= 1.0:
                    unit_name, unit_scale = name, scale
                    break

        scaled = v / unit_scale
        # Use a stable, parseable representation (roundtrip-ish for float).
        num_s = format(scaled, '.15g')
        return f'{sign}{num_s} {unit_name}'

    _PARSE_RE: ta.ClassVar[re.Pattern[str]] = re.compile(
        r'^\s*([+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?)\s*([A-Za-z]+)\s*$',
    )

    @classmethod
    def parse(cls, s: str) -> 'Seconds':
        m = cls._PARSE_RE.match(s)
        if not m:
            raise ValueError(f'Invalid Seconds string: {s!r}')
        num_s, unit = m.group(1), m.group(2)

        num = float(num_s)
        try:
            scale = cls._UNITS_MAP[unit]
        except KeyError:
            raise ValueError(f'Unknown time unit {unit!r}') from None

        return cls(num * scale)

    # operators

    def __hash__(self) -> int:
        return hash(self._v)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Seconds):
            return NotImplemented
        return self._v == other._v

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Seconds):
            return NotImplemented
        return self._v != other._v

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Seconds):
            return NotImplemented
        return self._v < other._v

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Seconds):
            return NotImplemented
        return self._v <= other._v

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Seconds):
            return NotImplemented
        return self._v > other._v

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Seconds):
            return NotImplemented
        return self._v >= other._v

    def _coerce_other(self, other: object) -> float:
        if isinstance(other, Seconds):
            return other._v  # Noqa
        elif _is_exact_int_or_float(other):
            return float(other)  # type: ignore[arg-type]
        else:
            raise TypeError(f'Seconds can only add/sub Seconds or exact int/float, got {type(other).__name__}')

    def __add__(self, other: object) -> 'Seconds':
        return Seconds(self._v + self._coerce_other(other))

    def __radd__(self, other: object) -> 'Seconds':
        return self.__add__(other)

    def __sub__(self, other: object) -> 'Seconds':
        return Seconds(self._v - self._coerce_other(other))

    def __rsub__(self, other: object) -> 'Seconds':
        if isinstance(other, Seconds):
            return Seconds(other._v - self._v)
        if _is_exact_int_or_float(other):
            return Seconds(float(other) - self._v)  # type: ignore[arg-type]
        return NotImplemented

    def __mul__(self, other: object) -> 'Seconds':
        if _is_exact_int_or_float(other):
            return Seconds(self._v * float(other))  # type: ignore[arg-type]
        raise TypeError(f'Seconds can only be multiplied by exact int/float, got {type(other).__name__}')

    def __rmul__(self, other: object) -> 'Seconds':
        return self.__mul__(other)

    def __truediv__(self, other: object) -> 'Seconds':
        if _is_exact_int_or_float(other):
            other_f = float(other)  # type: ignore[arg-type]
            if other_f == 0.0:
                raise ZeroDivisionError
            return Seconds(self._v / other_f)
        raise TypeError(f'Seconds can only be divided by exact int/float, got {type(other).__name__}')

    def __pow__(self, other: object, modulo: object = None) -> 'Seconds':
        raise TypeError('Seconds exponentiation is nonsensical for this unit type')

    # timedelta interop

    def to_timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self._v)

    @classmethod
    def from_timedelta(cls, td: datetime.timedelta) -> 'Seconds':
        return cls(td.total_seconds())


S: ta.Final[Seconds] = Seconds(1.0)
MS: ta.Final[Seconds] = Seconds(1e-3)
US: ta.Final[Seconds] = Seconds(1e-6)
NS: ta.Final[Seconds] = Seconds(1e-9)

M: ta.Final[Seconds] = 60 * S
H: ta.Final[Seconds] = 60 * M
D: ta.Final[Seconds] = 24 * H



import dataclasses as dc
import json
import os.path
import re
import shutil
import subprocess
import sys
import typing as ta

from .. import check
from .. import lang


if ta.TYPE_CHECKING:
    import plistlib

    from ..docker import all as docker

else:
    plistlib = lang.proxy_import('plistlib')

    docker = lang.proxy_import('..docker.all')


##


PYCHARM_HOSTED_ENV_VAR = 'PYCHARM_HOSTED'


def is_pycharm_hosted() -> bool:
    return PYCHARM_HOSTED_ENV_VAR in os.environ


##


DARWIN_PYCHARM_HOME = '/Applications/PyCharm.app'


def read_darwin_pycharm_info_plist() -> ta.Mapping[str, ta.Any] | None:
    plist_file = os.path.join(DARWIN_PYCHARM_HOME, 'Contents', 'Info.plist')
    if not os.path.isfile(plist_file):
        return None

    with open(plist_file, 'rb') as f:
        root = plistlib.load(f)

    return root


#


UBUNTU_PYCHARM_HOME = '/snap/pycharm-professional/current'


def read_ubuntu_pycharm_product_info() -> ta.Mapping[str, ta.Any] | None:
    json_file = os.path.join(UBUNTU_PYCHARM_HOME, 'product-info.json')
    if not os.path.isfile(json_file):
        return None

    with open(json_file) as f:
        root = json.load(f)

    return root


#


@lang.cached_function
def get_pycharm_version() -> str | None:
    if sys.platform == 'darwin':
        plist = read_darwin_pycharm_info_plist()
        if plist is None:
            return None

        ver = check.non_empty_str(plist['CFBundleVersion'])
        check.state(ver.startswith('PY-'))
        return ver[3:]

    elif sys.platform == 'linux':
        if shutil.which('lsb_release') is not None:
            lsb_id = subprocess.check_output(['lsb_release', '-is']).decode().strip()
            if lsb_id == 'Ubuntu':
                pi = read_ubuntu_pycharm_product_info()
                if pi is not None:
                    ver = check.non_empty_str(pi['buildNumber'])
                    return ver

        return None

    else:
        return None


##


def _import_pydevd_pycharm(*, version: str | None = None) -> ta.Any:
    if (
            'pydevd_pycharm' in sys.modules or
            (version is None and lang.can_import('pydevd_pycharm'))
    ):
        # Can't unload, nothing we can do
        import pydevd_pycharm  # noqa
        return pydevd_pycharm

    proc = subprocess.run([  # noqa
        sys.executable,
        '-m', 'pip',
        'show',
        'pydevd_pycharm',
    ], stdout=subprocess.PIPE)

    if not proc.returncode:
        info = {
            k: v.strip()
            for l in proc.stdout.decode().splitlines()
            if (s := l.strip())
            for k, _, v in [s.partition(':')]
        }

        installed_version = info['Version']
        if installed_version == version:
            import pydevd_pycharm  # noqa
            return pydevd_pycharm

        subprocess.check_call([
            sys.executable,
            '-m', 'pip',
            'uninstall', '-y',
            'pydevd_pycharm',
        ])

    subprocess.check_call([
        sys.executable,
        '-m', 'pip',
        'install',
        'pydevd_pycharm' + (f'=={version}' if version is not None else ''),
    ])

    import pydevd_pycharm  # noqa
    return pydevd_pycharm


@dc.dataclass(frozen=True)
class PycharmRemoteDebugger:
    host: str | None
    port: int
    version: str | None = None

    def __str__(self) -> str:
        return ''.join([
            f'@{self.version}:' if self.version is not None else '',
            f'{self.host}:' if self.host is not None else '',
            str(self.port),
        ])

    @classmethod
    def parse(cls, s: str) -> 'PycharmRemoteDebugger':
        if (m := re.fullmatch(r'(@(?P<version>[^:]+):)?((?P<host>[^:]+):)?(?P<port>\d+)', s)) is None:
            raise ValueError(s)
        gd = m.groupdict()
        return cls(
            host=gd.get('host'),
            port=int(gd['port']),
            version=gd.get('version'),
        )


def pycharm_remote_debugger_attach(prd: PycharmRemoteDebugger) -> None:
    host = prd.host
    if host is None:
        if (
                sys.platform == 'linux' and
                docker.is_likely_in_docker() and
                docker.get_docker_host_platform() == 'darwin'
        ):
            host = docker.DOCKER_FOR_MAC_HOSTNAME
        else:
            host = 'localhost'

    version = prd.version
    if version is None and host in ('localhost', '127.0.0.1'):
        version = get_pycharm_version()

    if ta.TYPE_CHECKING:
        import pydevd_pycharm  # noqa
    else:
        pydevd_pycharm = _import_pydevd_pycharm(version=version)

    import inspect
    st_sig = inspect.signature(pydevd_pycharm.settrace)

    kw: dict = {}
    if 'stdoutToServer' in st_sig.parameters:
        kw.update(
            stdoutToServer=True,
            stderrToServer=True,
        )
    else:
        kw.update(
            stdout_to_server=True,
            stderr_to_server=True,
        )

    pydevd_pycharm.settrace(
        host,
        port=prd.port,
        **kw,
    )



# ruff: noqa: UP045
# @omlish-lite
import asyncio
import typing as ta

from ..semaphores import AsyncliteSemaphore
from ..semaphores import AsyncliteSemaphores
from .base import AsyncioAsyncliteApi
from .base import AsyncioAsyncliteObject


##


class AsyncioAsyncliteSemaphore(AsyncliteSemaphore, AsyncioAsyncliteObject):
    def __init__(self, u: asyncio.Semaphore) -> None:
        super().__init__()

        self._u = u

    async def acquire(self, *, timeout: ta.Optional[float] = None) -> None:
        await self._wait_for(self._u.acquire(), timeout=timeout)

    def acquire_nowait(self) -> bool:
        # asyncio.Semaphore doesn't have acquire_nowait, so we check if we can acquire
        if self._u.locked():
            return False

        # Manually decrement the internal counter
        self._u._value -= 1  # noqa
        return True

    def release(self) -> None:
        self._u.release()


class AsyncioAsyncliteSemaphores(AsyncliteSemaphores, AsyncioAsyncliteApi):
    def make_semaphore(self, value: int = 1) -> AsyncliteSemaphore:
        return AsyncioAsyncliteSemaphore(asyncio.Semaphore(value))



import abc
import concurrent.futures as cf
import functools
import queue
import threading
import typing as ta

from .. import check
from .. import lang


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


class ThreadRunnerClosedError(Exception):
    pass


class ThreadRunner(lang.Abstract):
    @abc.abstractmethod
    def run(self, fn: ta.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        raise NotImplementedError

    DEFAULT_CLOSE_TIMEOUT: ta.ClassVar[float | None] = 10.

    @abc.abstractmethod
    def close(self, *, wait: bool = False, timeout: float | None = DEFAULT_CLOSE_TIMEOUT) -> None:
        raise NotImplementedError

    #

    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(wait=True)


##


class BaseThreadRunner(ThreadRunner, lang.Abstract):
    def __init__(
            self,
            *,
            name: str | None = None,
    ) -> None:
        super().__init__()

        self._name = name

        self._lock = threading.Lock()

        self._thread: threading.Thread | None = None
        self._closed = False

    @property
    def name(self) -> str | None:
        return self._name

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}({f"name={self._name!r}" if self._name is not None else ""})'

    class _Stop:
        pass

    def _ensure_thread(self) -> threading.Thread:
        if self._thread is None:
            t = threading.Thread(
                target=self._thread_main,
                name=f'{self!r}-worker',
                daemon=True,
            )
            t.start()
            self._thread = t
        return self._thread

    @abc.abstractmethod
    def _thread_main(self) -> None:
        raise NotImplementedError

    def run(self, fn: ta.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        if args or kwargs:
            fn = functools.partial(fn, *args, **kwargs)

        return self._run(fn)

    @abc.abstractmethod
    def _run(self, fn: ta.Callable[[], T]) -> T:
        raise NotImplementedError


##


class SingleThreadRunner(BaseThreadRunner):
    def __init__(
            self,
            *,
            name: str | None = None,
    ) -> None:
        super().__init__(name=name)

        self._req_cv = threading.Condition(self._lock)
        self._res_cv = threading.Condition(self._lock)

        self._req: ta.Callable[[], ta.Any] | SingleThreadRunner._Stop | None = None
        self._res: SingleThreadRunner._Result | None = None

        self._run_lock = threading.Lock()

    class _Result(ta.NamedTuple):
        v: ta.Any
        e: BaseException | None

    def _thread_main(self) -> None:
        while True:
            with self._req_cv:
                while self._req is None:
                    self._req_cv.wait()

                check.none(self._res)
                req = check.not_none(self._req)
                self._req = None

            if isinstance(req, self._Stop):
                with self._res_cv:
                    self._res = self._Result(req, None)
                    self._res_cv.notify()
                return

            v: ta.Any = None
            e: BaseException | None = None
            try:
                v = req()  # type: ignore[operator]
            except BaseException as e2:  # noqa
                e = e2

            res = self._Result(v, e)
            del v, e

            with self._res_cv:
                check.none(self._res)
                self._res = res
                self._res_cv.notify()

    def _run(self, fn: ta.Callable[[], T]) -> T:
        with self._run_lock:
            with self._lock:
                check.state(not self._closed, ThreadRunnerClosedError)
                check.none(self._req)
                check.none(self._res)

                self._ensure_thread()

                self._req = fn
                self._req_cv.notify()

                while self._res is None:
                    self._res_cv.wait()

                res = check.not_none(self._res)

                self._res = None

        if res.e is not None:
            raise res.e
        return res.v

    def close(self, *, wait: bool = False, timeout: float | None = BaseThreadRunner.DEFAULT_CLOSE_TIMEOUT) -> None:
        with self._run_lock:
            t = self._thread
            if t is None:
                self._closed = True
                return

            with self._lock:
                if not self._closed:
                    self._closed = True
                    check.none(self._req)
                    check.none(self._res)

                    self._req = self._Stop()
                    self._req_cv.notify()

                    while self._res is None:
                        self._res_cv.wait()

                    res = check.not_none(self._res)
                    check.isinstance(res.v, self._Stop)

                    self._req = None
                    self._res = None

        if wait:
            t.join(timeout)
            if t.is_alive():
                raise TimeoutError


##


class QueueThreadRunner(BaseThreadRunner):
    def __init__(
            self,
            *,
            name: str | None = None,
    ) -> None:
        super().__init__(name=name)

        self._queue: queue.Queue[QueueThreadRunner._Call | QueueThreadRunner._Stop] = queue.Queue()

    class _Call(ta.NamedTuple):
        fn: ta.Callable[[], ta.Any]
        fut: cf.Future[ta.Any]

    def _thread_main(self) -> None:
        while True:
            o = self._queue.get()
            try:
                if isinstance(o, self._Stop):
                    return

                try:
                    o.fut.set_result(o.fn())
                except BaseException as e:  # noqa
                    o.fut.set_exception(e)

            finally:
                self._queue.task_done()

    def _run(self, fn: ta.Callable[[], T]) -> T:
        fut: cf.Future[T] = cf.Future()

        with self._lock:
            if self._closed:
                raise ThreadRunnerClosedError

            self._ensure_thread()
            self._queue.put(self._Call(
                fn,
                fut,
            ))

        return fut.result()

    @ta.override
    def close(self, *, wait: bool = False, timeout: float | None = BaseThreadRunner.DEFAULT_CLOSE_TIMEOUT) -> None:
        with self._lock:
            if self._closed:
                t = self._thread
            else:
                self._closed = True
                t = self._thread
                if t is not None:
                    self._queue.put(self._Stop())

        if t is None:
            return

        if wait:
            t.join(timeout)
            if t.is_alive():
                raise TimeoutError



"""
TODO:
 - untangle compilation from here
 - populate linecache
"""
import abc
import dataclasses as dc
import sys
import typing as ta

from .... import check
from .... import lang
from ....logs import all as logs
from ..processing.base import ProcessingContext
from ..processing.base import ProcessingOption
from ..processing.base import Processor
from ..processing.phases import ProcessorPhase
from ..processing.registry import register_processor_type
from .base import Plan
from .compilation import OpCompiler
from .execution import OpExecutor
from .globals import FN_GLOBALS
from .globals import FnGlobal
from .idents import CLS_IDENT
from .ops import Op
from .ops import OpRef
from .ops import OpRefMap
from .plans import Plans
from .registry import all_generator_types
from .registry import generator_type_for_plan_type


log = logs.get_module_logger(globals())


##


@dc.dataclass(frozen=True)
class PlanOnly(ProcessingOption):
    b: bool


@dc.dataclass(frozen=True, kw_only=True)
class Verbosity(ProcessingOption):
    warn: bool = False
    debug: bool = False


class CompileCallback(ta.Protocol):
    def __call__(
            self,
            ctx: ProcessingContext,
            prepared: 'GeneratorProcessor.Prepared',
            comp: OpCompiler.CompileResult,
    ) -> None:
        ...


@dc.dataclass(frozen=True, kw_only=True)
class Codegen(ProcessingOption):
    style: ta.Literal['jit', 'aot'] = 'jit'

    force: bool = False
    callback: CompileCallback | None = None


##


@register_processor_type(phase=ProcessorPhase.GENERATION)
class GeneratorProcessor(Processor):
    PROCESS_FN_NAME: ta.ClassVar[str] = '_process_dataclass'

    class Mode(lang.Abstract):
        @abc.abstractmethod
        def _process(self, gp: 'GeneratorProcessor', cls: type) -> None:
            raise NotImplementedError

    class ExecutorMode(Mode):
        def _process(self, gp: 'GeneratorProcessor', cls: type) -> None:
            opx = OpExecutor(
                cls,
                gp.prepare().ref_map,
            )

            for op in gp.ops():
                opx.execute(op)

    class CompilerMode(Mode):
        def __init__(
                self,
                *,
                codegen: Codegen,
        ) -> None:
            super().__init__()

            self._codegen = codegen

        @classmethod
        def build_standard_kwargs(cls, dc_cls: type) -> dict[str, ta.Any]:
            kw: dict = {CLS_IDENT: dc_cls}
            kw.update({
                k.ident: v.value
                for k, v in FN_GLOBALS.items()
            })
            return kw

        def _process(self, gp: 'GeneratorProcessor', cls: type) -> None:
            style: OpCompiler.Style = {
                'jit': OpCompiler.JitStyle,
                'aot': OpCompiler.AotStyle,
            }[self._codegen.style]()

            compiler = OpCompiler(style)

            comp = compiler.compile(
                GeneratorProcessor.PROCESS_FN_NAME,
                gp.ops(),
            )

            comp_src = '\n'.join([
                *comp.hdr_lines,
                *(['', ''] if comp.hdr_lines else []),
                *comp.fn_lines,
            ])

            if (vo := gp._ctx.option(Verbosity)) is not None and vo.debug:  # noqa
                print(gp.prepare().plans.repr(), file=sys.stderr)
                print(file=sys.stderr)
                print(comp_src, file=sys.stderr)
                print(file=sys.stderr)

            ns: dict = {}
            ns.update(compiler.style.globals_ns())  # noqa

            exec(comp_src, ns)
            o_fn = ns[GeneratorProcessor.PROCESS_FN_NAME]

            if cls.__module__ in sys.modules:
                gl = sys.modules[cls.__module__].__dict__
            else:
                gl = {}

            # TODO: comment why lol
            fn = lang.new_function(**{
                **lang.new_function_kwargs(o_fn),
                **dict(
                    globals=gl,
                ),
            })

            kw = self.build_standard_kwargs(cls)

            orm = gp.prepare().ref_map
            for r in comp.refs:
                if isinstance(r, OpRef):
                    kw[r.ident()] = orm[r]
                elif isinstance(r, FnGlobal):
                    pass
                else:
                    raise TypeError(r)

            fn(**kw)

            if (cg := self._codegen) is not None and (cb := cg.callback) is not None:
                cb(  # noqa
                    gp._ctx,  # noqa
                    gp.prepare(),
                    comp,
                )

    #

    @dc.dataclass(frozen=True)
    class Prepared:
        plans: Plans
        ref_map: OpRefMap

        def __post_init__(self) -> None:
            hash(self.plans)

    @lang.cached_function(no_wrapper_update=True)
    def prepare(self) -> Prepared:
        gs = [g_ty() for g_ty in all_generator_types()]

        pll: list[Plan] = []
        orm: dict[OpRef, ta.Any] = {}
        for g in gs:
            if (pr := g.plan(self._ctx)) is None:
                continue

            for k, v in (pr.ref_map or {}).items():
                if k in orm:
                    check.equal(orm[k], v)
                else:
                    orm[k] = v

            pll.append(pr.plan)

        plans = Plans(tuple(pll))

        return self.Prepared(
            plans,
            orm,
        )

    @lang.cached_function(no_wrapper_update=True)
    def ops(self) -> ta.Sequence[Op]:
        prepared = self.prepare()

        ops: list[Op] = []
        for pl in prepared.plans:
            g = generator_type_for_plan_type(type(pl))()
            ops.extend(g.generate(pl))

        return ops

    #

    def _process_from_codegen(self, cls: type) -> bool:
        cg_pkg = check.not_none(self._ctx.pkg_cfg.pkg)
        cg_mod_spec = f'{cg_pkg}._dataclasses'

        try:
            __import__(cg_mod_spec)
        except ImportError:
            if (vo := self._ctx.option(Verbosity)) is not None and vo.warn:  # noqa
                log.warning(lambda: f'Codegen module missing for {cls.__module__}.{cls.__qualname__} at {cg_mod_spec}')
            return False

        cg_mod = sys.modules[cg_mod_spec]
        cg_fn_reg = cg_mod.REGISTRY

        #

        prep = self.prepare()
        prep_plan_repr = prep.plans.repr()

        #

        try:
            cg_reg_item = cg_fn_reg[prep_plan_repr]
        except KeyError:
            if (vo := self._ctx.option(Verbosity)) is not None and vo.warn:  # noqa
                log.warning(lambda: f'Codegen missing for {cls.__module__}.{cls.__qualname__} in {cg_mod_spec}')
            return False

        cg_kw, cg_fn = cg_reg_item

        #

        ref_map = {
            ref.ident(): v
            for ref, v in prep.ref_map.items()
        }

        #

        fn_kw = {
            **GeneratorProcessor.CompilerMode.build_standard_kwargs(cls),
            **{k: ref_map[k] for k in cg_kw['op_ref_idents']},
        }

        fn = cg_fn()
        fn(**fn_kw)

        return True

    #

    def process(self, cls: type) -> type:
        if (po := self._ctx.option(PlanOnly)) is not None and po.b:
            self.prepare()
            return cls

        cg = self._ctx.option(Codegen)

        if not (cg is not None and cg.force) and self._ctx.pkg_cfg.cfg.codegen:
            if self._process_from_codegen(cls):
                return cls

        mode: GeneratorProcessor.Mode
        if cg is not None:  # noqa
            mode = GeneratorProcessor.CompilerMode(codegen=cg)
        else:
            mode = GeneratorProcessor.ExecutorMode()

        mode._process(self, cls)  # noqa
        return cls



# ruff: noqa: UP043 UP045
# @omlish-lite
import dataclasses as dc

from .asyncs import AsyncHttpClient
from .asyncs import AsyncStreamHttpClientResponse
from .base import HttpClientContext
from .base import HttpClientRequest
from .sync import HttpClient
from .sync import StreamHttpClientResponse


##


class SyncAsyncHttpClient(AsyncHttpClient):
    def __init__(self, client: HttpClient) -> None:
        super().__init__()

        self._client = client

    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        ul: StreamHttpClientResponse

        async def read1(self, n: int = -1, /) -> bytes:
            return self.ul.stream.read1(n)

        async def read(self, n: int = -1, /) -> bytes:
            return self.ul.stream.read(n)

        async def close(self) -> None:
            self.ul.close()

    async def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> AsyncStreamHttpClientResponse:
        resp = self._client.stream_request(req, context=ctx)
        return AsyncStreamHttpClientResponse(
            status=resp.status,
            headers=resp.headers,
            request=req,
            underlying=resp,
            **(dict(  # type: ignore
                _stream=(adapter := self._StreamAdapter(resp)),
                _closer=adapter.close,
            ) if resp.has_data else {}),
        )



# ruff: noqa: I001
from ... import lang as _lang


with _lang.auto_proxy_init(globals()):
    from .api import (  # noqa
        SqlalchemyApiWrapper,

        SqlalchemyApiRows,
        SqlalchemyApiConn,
        SqlalchemyApiDb,
        SqlalchemyApiAdapter,

        api_adapt,
    )


from .exprs import (  # noqa
    paren,
)



# @omlish-lite


##


class ByteStreamBufferError(Exception):
    pass


#


class NeedMoreDataByteStreamBufferError(ByteStreamBufferError):
    """
    Raised when an operation cannot complete because insufficient bytes are currently buffered.

    This is intentionally distinct from EOF: it means "try again after feeding more bytes".
    """


#


class LimitByteStreamBufferError(ValueError, ByteStreamBufferError):
    """
    Base class for buffer/framing limit violations.

    Subclasses inherit from ValueError so existing tests expecting ValueError continue to pass.
    """


class BufferTooLargeByteStreamBufferError(LimitByteStreamBufferError):
    """
    Buffered data exceeded a configured cap without finding a boundary that would allow progress.

    Typically indicates an unframed stream, a missing delimiter, or an upstream not enforcing limits.
    """


class FrameTooLargeByteStreamBufferError(LimitByteStreamBufferError):
    """A single decoded frame (payload before its boundary delimiter/length) exceeded a configured max size."""


#


class StateByteStreamBufferError(RuntimeError, ByteStreamBufferError):
    """
    Base class for invalid buffer state transitions (e.g., coalescing while a reservation is outstanding).

    Subclasses inherit from RuntimeError so existing tests expecting RuntimeError continue to pass.
    """


class OutstandingReserveByteStreamBufferError(StateByteStreamBufferError):
    """A reserve() is outstanding; an operation requiring stable storage cannot proceed."""


class NoOutstandingReserveByteStreamBufferError(StateByteStreamBufferError):
    """commit() was called without a preceding reserve()."""



# @omlish-lite
# ruff: noqa: UP006 UP007 UP043 UP045
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import typing as ta

from ....lite.check import check
from ..io import CoroHttpIo
from .errors import CoroHttpClientErrors


#


class CoroHttpClientStatusLine(ta.NamedTuple):
    version: str
    status: int
    reason: str

    @classmethod
    def read(cls) -> ta.Generator[CoroHttpIo.Io, ta.Optional[bytes], 'CoroHttpClientStatusLine']:
        line = str(check.isinstance((yield CoroHttpIo.ReadLineIo(CoroHttpIo.MAX_LINE + 1)), bytes), 'iso-8859-1')
        if len(line) > CoroHttpIo.MAX_LINE:
            raise CoroHttpClientErrors.LineTooLongError(CoroHttpClientErrors.LineTooLongError.LineType.STATUS)
        if not line:
            # Presumably, the server closed the connection before sending a valid response.
            raise CoroHttpClientErrors.RemoteDisconnectedError('Remote end closed connection without response')

        version = ''
        reason = ''
        status_str = ''
        try:
            version, status_str, reason = line.split(None, 2)
        except ValueError:
            try:
                version, status_str = line.split(None, 1)
            except ValueError:
                # empty version will cause next test to fail.
                pass

        if not version.startswith('HTTP/'):
            raise CoroHttpClientErrors.BadStatusLineError(line)

        # The status code is a three-digit number
        try:
            status = int(status_str)
        except ValueError:
            raise CoroHttpClientErrors.BadStatusLineError(line) from None

        if status < 100 or status > 999:
            raise CoroHttpClientErrors.BadStatusLineError(line)

        return cls(version, status, reason)



import typing as ta

from ... import cached
from ... import collections as col
from ... import dataclasses as dc
from .api import DEFAULT_FIELD_OPTIONS
from .api import FieldOptions


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldInfo:
    """
    Computed field information - derived from FieldMetadata + dataclass introspection.

    This is purely derived/computed data, not configuration. The metadata field contains the final merged configuration.
    """

    name: str
    type: ta.Any

    marshal_name: str | None
    unmarshal_names: ta.Sequence[str]

    options: FieldOptions = DEFAULT_FIELD_OPTIONS


@dc.dataclass(frozen=True)
class FieldInfos:
    """Collection of field infos with convenient lookups."""

    lst: ta.Sequence[FieldInfo]

    def __iter__(self) -> ta.Iterator[FieldInfo]:
        return iter(self.lst)

    def __len__(self) -> int:
        return len(self.lst)

    @cached.property
    @dc.init
    def by_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((fi.name, fi) for fi in self), strict=True)

    @cached.property
    @dc.init
    def by_marshal_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((fi.marshal_name, fi) for fi in self if fi.marshal_name is not None), strict=True)

    @cached.property
    @dc.init
    def by_unmarshal_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((n, fi) for fi in self for n in fi.unmarshal_names), strict=True)



import dataclasses as dc
import typing as ta

from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.registry import register_generator_type
from ..processing.base import ProcessingContext
from .fields import InstanceFields


##


@dc.dataclass(frozen=True)
class EqPlan(Plan):
    fields: tuple[str, ...]


@register_generator_type(EqPlan)
class EqGenerator(Generator[EqPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[EqPlan] | None:
        if not ctx.cs.eq or '__eq__' in ctx.cls.__dict__:
            return None

        return PlanResult(EqPlan(
            tuple(f.name for f in ctx[InstanceFields] if f.compare),
        ))

    def generate(self, pl: EqPlan) -> ta.Iterable[Op]:
        ret_lines: list[str]
        if pl.fields:
            ret_lines = [
                f'    return (',
                *[
                    f'        self.{a} == other.{a}{" and" if i < len(pl.fields) - 1 else ""}'
                    for i, a in enumerate(pl.fields)
                ],
                f'    )',
            ]
        else:
            ret_lines = [
                f'    return True',
            ]

        return [
            AddMethodOp(
                '__eq__',
                '\n'.join([
                    f'def __eq__(self, other):',
                    f'    if self is other:',
                    f'        return True',
                    f'    if self.__class__ is not other.__class__:',
                    f'        return NotImplemented',
                    *ret_lines,
                ]),
            ),
        ]



# ruff: noqa: UP006 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ....lite.check import check
from ...streams.segmented import SegmentedByteStreamBuffer
from ...streams.types import MutableByteStreamBuffer
from ...streams.utils import ByteStreamBuffers
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineHandlerNotification
from ..core import IoPipelineHandlerNotifications
from ..core import IoPipelineMessages
from ..flow.types import IoPipelineFlow
from ..flow.types import IoPipelineFlowMessages
from .buffering import OutboundBytesBufferingIoPipelineHandler


##


@ta.final
class OutboundBytesBufferIoPipelineHandler(OutboundBytesBufferingIoPipelineHandler):
    @ta.final
    @dc.dataclass(frozen=True)
    class Config:
        flush_threshold: ta.Optional[int] = 8 * 1024
        max_buffer_size: ta.Optional[int] = None
        buffer_chunk_size: int = 64 * 1024

        DEFAULT: ta.ClassVar['OutboundBytesBufferIoPipelineHandler.Config']

    Config.DEFAULT = Config()

    def __init__(self, config: ta.Optional[Config] = None) -> None:
        super().__init__()

        if config is None:
            config = self.Config.DEFAULT
        self._config = config

    #

    _buf: ta.Optional[MutableByteStreamBuffer] = None

    def _new_buf(self) -> MutableByteStreamBuffer:
        return SegmentedByteStreamBuffer(
            max_size=self._config.max_buffer_size,
            chunk_size=self._config.buffer_chunk_size,
        )

    def outbound_buffered_bytes(self) -> ta.Optional[int]:
        if (buf := self._buf) is None:
            return 0
        return len(buf)

    #

    def notify(self, ctx: IoPipelineHandlerContext, no: IoPipelineHandlerNotification) -> None:
        if isinstance(no, IoPipelineHandlerNotifications.Added):
            # This handler *requires* flow
            check.not_none(no.ctx.services.find(IoPipelineFlow))

    #

    def _flush(self, ctx: IoPipelineHandlerContext) -> None:
        if (buf := self._buf) is None or len(buf) == 0:
            return

        # Collect all buffered segments and feed them out
        for seg in buf.segments():
            ctx.feed_out(seg)

        # Reset buffer
        self._buf = None

    _FLUSH_AND_FEED_OUT_TYPES: ta.ClassVar[ta.Tuple[type, ...]] = (
        IoPipelineFlowMessages.FlushOutput,
        IoPipelineMessages.FinalOutput,
    )

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, self._FLUSH_AND_FEED_OUT_TYPES):
            self._flush(ctx)
            ctx.feed_out(msg)

        elif ByteStreamBuffers.can_bytes(msg):
            # Buffer the bytes
            if (buf := self._buf) is None:
                buf = self._buf = self._new_buf()

            for seg in ByteStreamBuffers.iter_segments(msg):
                if seg:
                    buf.write(seg)

            # Check if we should flush based on threshold
            if (
                (threshold := self._config.flush_threshold) is not None and
                len(buf) >= threshold
            ):
                self._flush(ctx)

        else:
            ctx.feed_out(msg)



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ..headers import HttpHeaders


##


@dc.dataclass()
class IoPipelineHttpBodyModeError(Exception):
    reason: str


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpBodyMode:
    mode: ta.Literal['empty', 'eof', 'cl', 'chunked']
    length: ta.Optional[int]

    @classmethod
    def select(
            cls,
            headers: HttpHeaders,
            *,
            if_length_missing: ta.Literal['empty', 'eof'],
    ) -> 'IoPipelineHttpBodyMode':
        if headers.contains_value('transfer-encoding', 'chunked', ignore_case=True):
            return cls('chunked', None)

        cl = headers.single.get('content-length')
        if not cl:
            return cls(if_length_missing, None)

        try:
            n = int(cl)
        except ValueError:
            raise IoPipelineHttpBodyModeError('bad Content-Length') from None

        if n < 0:
            raise IoPipelineHttpBodyModeError('bad Content-Length')

        if n == 0:
            return cls('empty', None)

        return cls('cl', n)



import typing as ta

from ... import dataclasses as dc
from .api import DEFAULT_FIELD_OPTIONS
from .api import FieldOptions
from .api import ObjectOptions


T = ta.TypeVar('T')


##


def with_field_options(**kwargs: ta.Any) -> dc.field_modifier:
    @dc.field_modifier
    def inner(f: dc.Field) -> dc.Field:
        existing = f.metadata.get(FieldOptions, DEFAULT_FIELD_OPTIONS)
        updated = dc.replace(existing, **kwargs)
        return dc.set_field_metadata(f, {FieldOptions: updated})

    return inner


def update_fields_options(
        fields: ta.Iterable[str] | None = None,
        **kwargs: ta.Any,
) -> ta.Callable[[type[T]], type[T]]:
    def inner(a: str, f: dc.Field) -> dc.Field:
        existing = f.metadata.get(FieldOptions, DEFAULT_FIELD_OPTIONS)
        updated = dc.replace(existing, **kwargs)
        return dc.set_field_metadata(f, {FieldOptions: updated})

    return dc.update_fields(inner, fields)


def update_object_options(
        cls: type | None = None,
        **kwargs: ta.Any,
):
    def inner(cls):
        return dc.append_class_metadata(cls, ObjectOptions(**kwargs))

    if cls is not None:
        inner(cls)
        return cls
    else:
        return inner



import types
import typing as ta
import weakref


T = ta.TypeVar('T')


##


_STATIC_VALUES: ta.MutableMapping[types.CodeType, ta.Any] = weakref.WeakKeyDictionary()


def static(fn: ta.Callable[[], T]) -> T:
    """
    An 'inline' nullary cache, useful for late initialization of things that would be inconvenient to have to move
    outside of their point of use to a more proper for shared cached objects like a class or module body.

    Cached contents are effectively immortal (more specifically tied to the lifetime of the passed code object). Users
    must be careful to ensure to local state is unintentionally closed over.

    There is intentionally no locking performed - before population, any number of threads may simultaneously call the
    function and they will all execute like normal, but afterward only (an unspecified) one will remain in the cache -
    this mechanism alone does not provide 'exactly-once' semantics.

    Primarily intended for the inline construction of immutable objects - particularly things like query builders. For
    example:

    >>> sql.api.query_opt_one(db, lang.static(lambda: Q.select(
    >>>     [Q.i.value],
    >>>     Q.n.states,
    >>>     Q.eq(Q.n.key, Q.p.key),
    >>> )), {
    >>>     Q.p.key: key,
    >>> })

    As the cache will 'stabilize' to a single value after any possible concurrent initialization the returned object is
    generally suitable for use as key in separate caches (like query compilation caches).
    """

    try:
        return _STATIC_VALUES[fn.__code__]  # noqa
    except KeyError:
        pass

    _STATIC_VALUES[fn.__code__] = v = fn()  # noqa
    return v



import abc
import typing as ta

from ... import check
from ... import lang
from .adapters import HasAdapter
from .queries import Queryable


if ta.TYPE_CHECKING:
    from .core import AsyncRows
    from .core import Rows


##


class AnyQuerier(HasAdapter, lang.Abstract):
    pass


class Querier(AnyQuerier, lang.Abstract):
    @abc.abstractmethod
    def query(self, query: Queryable) -> ta.ContextManager['Rows']:  # ta.Raises[QueryError]
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass_except_nameerror(cls, lambda: AsyncQuerier)


class AsyncQuerier(AnyQuerier, lang.Abstract):
    @abc.abstractmethod
    def query(self, query: Queryable) -> ta.AsyncContextManager['AsyncRows']:  # ta.Raises[QueryError]
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass(cls, Querier)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ....io.pipelines.core import IoPipelineHandler
from ....io.pipelines.core import IoPipelineHandlerContext
from ....io.pipelines.core import IoPipelineMessages
from ....io.pipelines.flow.types import IoPipelineFlow
from ....io.pipelines.flow.types import IoPipelineFlowMessages
from ....lite.check import check
from ....lite.namespaces import NamespaceClass
from ..requests import FullIoPipelineHttpRequest
from ..responses import FullIoPipelineHttpResponse
from ..responses import IoPipelineHttpResponseEnd
from ..responses import IoPipelineHttpResponseObject
from .responses import IoPipelineHttpResponseAggregatorDecoder


##


class IoPipelineHttpClientMessages(NamespaceClass):
    @dc.dataclass(frozen=True)
    class Request(IoPipelineMessages.NeverOutbound):
        request: FullIoPipelineHttpRequest

        aggregate: ta.Union[bool, ta.Literal['unless_chunked'], None] = None

    @dc.dataclass(frozen=True)
    class Output(IoPipelineMessages.NeverInbound):
        msg: ta.Union[
            IoPipelineHttpResponseObject,
            IoPipelineMessages.FinalInput,
            'IoPipelineHttpClientMessages.Close',
        ]

        request: ta.Optional['IoPipelineHttpClientMessages.Request'] = None

    @dc.dataclass(frozen=True)
    class Close(IoPipelineMessages.NeverOutbound):
        pass


#


class IoPipelineHttpClientHandler(IoPipelineHandler):
    _request: ta.Optional[IoPipelineHttpClientMessages.Request] = None

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineHttpClientMessages.Request):
            check.none(self._request)
            self._request = msg

            if (ag := msg.aggregate) is not None:
                rad = check.not_none(ctx.pipeline.find_single_handler_of_type(IoPipelineHttpResponseAggregatorDecoder))
                rad.handler.set_enabled(ag)

            ctx.feed_out(msg.request)

            if (fc := ctx.services.find(IoPipelineFlow)) is not None:
                ctx.feed_out(IoPipelineFlowMessages.FlushOutput())

                if not fc.is_auto_read():
                    ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())

            return

        if isinstance(msg, IoPipelineHttpResponseObject):
            ctx.feed_out(IoPipelineHttpClientMessages.Output(msg, request=self._request))

            if isinstance(msg, (FullIoPipelineHttpResponse, IoPipelineHttpResponseEnd)):
                self._request = None

            return

        if isinstance(msg, (IoPipelineMessages.FinalInput, IoPipelineHttpClientMessages.Close)):
            ctx.feed_out(IoPipelineHttpClientMessages.Output(msg, request=self._request))

            self._request = None

            if isinstance(msg, IoPipelineMessages.FinalInput):
                ctx.feed_in(msg)

            ctx.feed_out(IoPipelineMessages.FinalOutput())

            return

        ctx.feed_in(msg)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import os.path

from ..lite.abstract import Abstract
from .mangle import mangle_path


##


class ShadowPaths(Abstract):
    @abc.abstractmethod
    def get_shadow_path(self, path: str) -> str:
        raise NotImplementedError


class ManglingShadowPaths(ShadowPaths):
    def __init__(
            self,
            root_path: str,
    ) -> None:  # noqa
        super().__init__()

        self._root_path = root_path

    def get_shadow_path(self, path: str) -> str:
        mangled = mangle_path(os.path.abspath(path))
        return os.path.abspath(os.path.join(self._root_path, mangled))



# ruff: noqa: UP006 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta


##


class IoPipelineError(Exception):
    pass


##


class UnhandleableIoPipelineError(IoPipelineError):
    pass


##
# state


class StateIoPipelineError(IoPipelineError):
    pass


class ContextInvalidatedIoPipelineError(StateIoPipelineError):
    pass


class SawInitialInputIoPipelineError(StateIoPipelineError):
    pass


class SawFinalInputIoPipelineError(StateIoPipelineError):
    pass


class SawFinalOutputIoPipelineError(StateIoPipelineError):
    pass


##
# messages


@dc.dataclass()
class MessageIoPipelineError(IoPipelineError):
    @ta.final
    @dc.dataclass(frozen=True)
    class Item:
        direction: ta.Literal['inbound', 'outbound']
        msg: ta.Any

        # _: dc.KW_ONLY

        last_seen: ta.Optional[ta.Any] = None

        def __repr__(self) -> str:
            return (
                f'{self.__class__.__name__}('
                f'{self.direction!r}'
                f', {self.msg!r}' +
                (f', last_seen={self.last_seen!r}' if self.last_seen is not None else '') +
                ')'
            )

    items: ta.Sequence[Item]

    @classmethod
    def new(
            cls,
            *,
            inbound: ta.Optional[ta.Sequence[ta.Any]] = None,
            inbound_with_last_seen: ta.Optional[ta.Sequence[ta.Tuple[ta.Any, ta.Any]]] = None,
            outbound: ta.Optional[ta.Sequence[ta.Any]] = None,
            outbound_with_last_seen: ta.Optional[ta.Sequence[ta.Tuple[ta.Any, ta.Any]]] = None,
    ) -> 'MessageIoPipelineError':
        items: ta.List[MessageIoPipelineError.Item] = []
        for msg in inbound or ():
            items.append(MessageIoPipelineError.Item('inbound', msg))
        for msg, last_seen in inbound_with_last_seen or ():
            items.append(MessageIoPipelineError.Item('inbound', msg, last_seen=last_seen))
        for msg in outbound or ():
            items.append(MessageIoPipelineError.Item('outbound', msg))
        for msg, last_seen in outbound_with_last_seen or ():
            items.append(MessageIoPipelineError.Item('outbound', msg, last_seen=last_seen))
        return cls(items)

    @classmethod
    def new_single(
            cls,
            direction: ta.Literal['inbound', 'outbound'],
            msg: ta.Any,
            *,
            last_seen: ta.Optional[ta.Any] = None,
    ) -> 'MessageIoPipelineError':
        return cls([
            MessageIoPipelineError.Item(
                direction,
                msg,
                last_seen=last_seen,
            ),
        ])

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.items!r})'


@dc.dataclass(repr=False)
class MessageNotPropagatedIoPipelineError(MessageIoPipelineError, UnhandleableIoPipelineError):
    pass


@dc.dataclass(repr=False)
class MessageReachedTerminalIoPipelineError(MessageIoPipelineError, UnhandleableIoPipelineError):
    pass


##
# misc (TODO: move/cleanup)


class DecodingIoPipelineError(IoPipelineError):
    pass


class IncompleteDecodingIoPipelineError(DecodingIoPipelineError):
    pass


class FlowControlValidationIoPipelineError(IoPipelineError):
    pass



# @omlish-lite
# ruff: noqa: PYI034 UP006 UP007 UP045
import threading
import time
import typing as ta

from ...lite.check import check
from ..addresses import SocketAndAddress
from .handlers import SocketServerHandler


##


class ThreadingSocketServerHandler:
    def __init__(
            self,
            handler: SocketServerHandler,
            *,
            shutdown_timeout: ta.Optional[float] = None,
    ) -> None:
        super().__init__()

        self._handler = handler
        self._shutdown_timeout = shutdown_timeout

        self._lock = threading.RLock()
        self._threads: ta.List[threading.Thread] = []
        self._is_shutdown = False

    @property
    def handler(self) -> SocketServerHandler:
        return self._handler

    #

    def __call__(self, conn: SocketAndAddress) -> None:
        self.handle(conn)

    def handle(self, conn: SocketAndAddress) -> None:
        with self._lock:
            check.state(not self._is_shutdown)

            self._reap()

            t = threading.Thread(
                target=self._handler,
                args=(conn,),
            )

            self._threads.append(t)

            t.start()

    #

    def _reap(self) -> None:
        with self._lock:
            self._threads[:] = (thread for thread in self._threads if thread.is_alive())

    def is_alive(self) -> bool:
        with self._lock:
            self._reap()

            return bool(self._threads)

    def join(self, timeout: ta.Optional[float] = None) -> None:
        if timeout is not None:
            deadline: ta.Optional[float] = time.time() + timeout
        else:
            deadline = None

        def calc_timeout() -> ta.Optional[float]:
            if deadline is None:
                return None

            tt = deadline - time.time()
            if tt <= 0:
                raise TimeoutError

            return tt

        if not (self._lock.acquire(timeout=calc_timeout() or -1)):
            raise TimeoutError

        try:
            self._reap()

            for t in self._threads:
                t.join(timeout=calc_timeout())

                if t.is_alive():
                    raise TimeoutError

        finally:
            self._lock.release()

    #

    class _NOT_SET:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    def shutdown(
            self,
            block: bool = False,
            timeout: ta.Union[float, None, ta.Type[_NOT_SET]] = _NOT_SET,
    ) -> None:
        self._is_shutdown = True

        if block:
            if timeout is self._NOT_SET:
                timeout = self._shutdown_timeout

            self.join(timeout=timeout)  # type: ignore

    #

    def __enter__(self) -> 'ThreadingSocketServerHandler':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()



# ruff: noqa: UP045
# @omlish-lite
import queue
import typing as ta

from ..queues import AsyncliteQueue
from ..queues import AsyncliteQueues
from .base import SyncAsyncliteApi
from .base import SyncAsyncliteObject


T = ta.TypeVar('T')


##


class SyncAsyncliteQueue(AsyncliteQueue[T], SyncAsyncliteObject):
    def __init__(self, u: 'queue.Queue[T]') -> None:
        super().__init__()

        self._u = u

    async def aclose(self) -> None:
        pass

    def qsize(self) -> int:
        return self._u.qsize()

    def empty(self) -> bool:
        return self._u.empty()

    def full(self) -> bool:
        return self._u.full()

    async def put(self, item: T, *, timeout: ta.Optional[float] = None) -> None:
        self._u.put(item, block=True, timeout=timeout)

    def put_nowait(self, item: T) -> None:
        self._u.put_nowait(item)

    async def get(self, *, timeout: ta.Optional[float] = None) -> T:
        return self._u.get(block=True, timeout=timeout)

    def get_nowait(self) -> T:
        return self._u.get_nowait()


class SyncAsyncliteQueues(AsyncliteQueues, SyncAsyncliteApi):
    def make_queue(self, maxsize: int = 0) -> AsyncliteQueue:
        return SyncAsyncliteQueue(queue.Queue(maxsize))



from .io import (  # noqa
    BufferedIOBase_AsyncIoProxy as BufferedIOBase,
    BufferedRandom_AsyncIoProxy as BufferedRandom,
    BufferedReader_AsyncIoProxy as BufferedReader,
    BufferedRWPair_AsyncIoProxy as BufferedRWPair,
    BufferedWriter_AsyncIoProxy as BufferedWriter,
    BytesIO_AsyncIoProxy as BytesIO,
    FileIO_AsyncIoProxy as FileIO,
    IOBase_AsyncIoProxy as IOBase,
    RawIOBase_AsyncIoProxy as RawIOBase,
    StringIO_AsyncIoProxy as StringIO,
    TextIOBase_AsyncIoProxy as TextIOBase,
    TextIOWrapper_AsyncIoProxy as TextIOWrapper,
)

from .proxier import (  # noqa
    AsyncIoProxier as Proxier,
)

from .proxy import (  # noqa
    AsyncIoProxy as Proxy,
    AsyncIoProxyRunner as Runer,
    AsyncIoProxyTarget as Target,
    async_io_proxy_cls_for as proxy_cls_for,
    async_io_proxy_fn as proxy_fn,
)

from .typing import (  # noqa
    TypingBinaryIO_AsyncIoProxy as TypingBinaryIO,
    TypingIO_AsyncIoProxy as TypingIO,
    TypingTextIO_AsyncIoProxy as TypingTextIO,
)



# ruff: noqa: UP006 UP007 UP037 UP045
# @omlish-lite
"""
TODO:
 - max size, simple backpressure?
"""
import collections
import typing as ta

from ....lite.abstract import Abstract
from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineHandlerFn
from ..core import IoPipelineMessages


##


class QueueIoPipelineHandler(IoPipelineHandler, Abstract):
    def __init__(
            self,
            *,
            filter: ta.Optional[IoPipelineHandlerFn[ta.Any, bool]] = None,  # noqa
            filter_type: ta.Optional[ta.Union[type, ta.Tuple[type, ...]]] = None,
            passthrough: ta.Union[bool, ta.Literal['must_propagate']] = 'must_propagate',
    ) -> None:
        super().__init__()

        self._filter = filter
        self._filter_type = filter_type
        self._passthrough = passthrough

        self._q: collections.deque[ta.Any] = collections.deque()

    def __repr__(self) -> str:
        return ''.join([
            f'{type(self).__name__}@{id(self):x}',
            f'<len={len(self._q)}>',
            '(',
            ', '.join([
                *([f'filter={self._filter!r}'] if self._filter is not None else []),
                *([f'filter_type={self._filter_type!r}'] if self._filter_type is not None else []),
                *([f'passthrough={self._passthrough!r}'] if self._passthrough else []),
            ]),
            ')',
        ])

    #

    def _append(self, msg: ta.Any) -> None:
        self._q.append(msg)

    def _popleft(self) -> ta.Any:
        return self._q.popleft()

    def poll(self) -> ta.Optional[ta.Any]:
        if not self._q:
            return None

        return self._popleft()

    def drain(self) -> ta.List[ta.Any]:
        out: ta.List[ta.Any] = []

        while self._q:
            out.append(self._popleft())

        return out

    #

    def _should_enqueue(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
        if self._filter is not None and not self._filter(ctx, msg):
            return False

        if self._filter_type is not None and not isinstance(msg, self._filter_type):
            return False

        return True

    def _should_passthrough(self, msg: ta.Any) -> bool:
        if isinstance(pt := self._passthrough, bool):
            return pt

        elif pt == 'must_propagate':
            return isinstance(msg, IoPipelineMessages.MustPropagate)

        else:
            raise RuntimeError(f'Unknown passthrough mode {self._passthrough!r} for {self!r}')

    def _handle(self, ctx: IoPipelineHandlerContext, msg: ta.Any, feed: ta.Callable[[ta.Any], None]) -> None:
        if not self._should_enqueue(ctx, msg):
            feed(msg)
            return

        self._append(msg)

        if self._should_passthrough(msg):
            feed(msg)


class InboundQueueIoPipelineHandler(QueueIoPipelineHandler):
    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._handle(ctx, msg, ctx.feed_in)


class OutboundQueueIoPipelineHandler(QueueIoPipelineHandler):
    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._handle(ctx, msg, ctx.feed_out)


class DuplexQueueIoPipelineHandler(
    InboundQueueIoPipelineHandler,
    OutboundQueueIoPipelineHandler,
):
    pass



import abc
import collections.abc
import enum
import functools
import typing as ta

from .. import check
from .. import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


ParamKey: ta.TypeAlias = str | int

SequencePreparedParams: ta.TypeAlias = ta.Sequence[ParamKey]
MappingPreparedParams: ta.TypeAlias = ta.Mapping[str, ParamKey]

PreparedParams: ta.TypeAlias = SequencePreparedParams | MappingPreparedParams


class ParamsPreparer(lang.Abstract):
    @abc.abstractmethod
    def add(self, k: ParamKey) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def prepare(self) -> PreparedParams:
        raise NotImplementedError


class SequenceParamsPreparer(ParamsPreparer, lang.Abstract):
    @abc.abstractmethod
    def prepare(self) -> SequencePreparedParams:
        raise NotImplementedError


class MappingParamsPreparer(ParamsPreparer, lang.Abstract):
    @abc.abstractmethod
    def prepare(self) -> MappingPreparedParams:
        raise NotImplementedError


##


class LinearParamsPreparer(SequenceParamsPreparer):
    def __init__(self, placeholder: str) -> None:
        super().__init__()

        self._placeholder = placeholder
        self._args: list[ParamKey] = []

    def add(self, k: ParamKey) -> str:
        self._args.append(k)
        return self._placeholder

    def prepare(self) -> SequencePreparedParams:
        return self._args


#


class NumericParamsPreparer(SequenceParamsPreparer):
    def __init__(self) -> None:
        super().__init__()

        self._args: list[ParamKey] = []
        self._str_by_key: dict[ParamKey, str] = {}
        self._pos_by_name: dict[str, int] = {}
        self._pos_by_id: dict[int, int] = {}

    def add(self, k: ParamKey) -> str:
        try:
            return self._str_by_key[k]
        except KeyError:
            pass

        pos = len(self._args)
        if isinstance(k, str):
            check.not_in(k, self._pos_by_name)
            self._pos_by_name[k] = pos
        elif isinstance(k, int):
            check.not_in(k, self._pos_by_id)
            self._pos_by_id[k] = pos
        else:
            raise TypeError(f'Invalid key type: {type(k)}')

        self._args.append(k)
        ret = self._str_by_key[k] = f':{pos + 1}'
        return ret

    def prepare(self) -> SequencePreparedParams:
        return self._args


#


class NamedParamsPreparer(MappingParamsPreparer):
    def __init__(
            self,
            render: ta.Callable[[str], str],
            *,
            anon: ta.Callable[[int], str] | None = None,
    ) -> None:
        super().__init__()

        self._render = render
        self._anon = anon if anon is not None else self.underscore_anon
        self._kwargs: dict[str, ParamKey] = {}
        self._str_by_key: dict[ParamKey, str] = {}
        self._kwargs_by_name: dict[str, str] = {}
        self._kwargs_by_id: dict[int, str] = {}

    @staticmethod
    def render_named(kwarg: str) -> str:
        return f':{kwarg}'

    @staticmethod
    def render_pyformat(kwarg: str) -> str:
        return f'%({kwarg})s'

    @staticmethod
    def underscore_anon(n: int) -> str:
        return f'_{n}'

    def add(self, k: ParamKey) -> str:
        try:
            return self._str_by_key[k]
        except KeyError:
            pass

        kwarg: str
        if isinstance(k, str):
            check.not_in(k, self._kwargs_by_name)
            kwarg = k  # TODO: collisions
            self._kwargs_by_name[k] = kwarg
        elif isinstance(k, int):
            check.not_in(k, self._kwargs_by_id)
            kwarg = self._anon(len(self._kwargs_by_id))  # TODO: collisions
            self._kwargs_by_id[k] = kwarg
        else:
            raise TypeError(f'Invalid key type: {type(k)}')

        self._kwargs[kwarg] = k
        ret = self._str_by_key[k] = self._render(kwarg)
        return ret

    def prepare(self) -> MappingPreparedParams:
        return self._kwargs


##


class ParamStyle(enum.Enum):
    QMARK = 'qmark'  # Question mark style, e.g. ...WHERE name=?
    FORMAT = 'format'  # ANSI C printf format codes, e.g. ...WHERE name=%s

    NUMERIC = 'numeric'  # Numeric, positional style, e.g. ...WHERE name=:1

    NAMED = 'named'  # Named style, e.g. ...WHERE name=:name
    PYFORMAT = 'pyformat'  # Python extended format codes, e.g. ...WHERE name=%(name)s


_PARAMS_PREPARER_FACTORIES_BY_STYLE: ta.Mapping[ParamStyle, ta.Callable[[], ParamsPreparer]] = {
    ParamStyle.QMARK: functools.partial(LinearParamsPreparer, '?'),
    ParamStyle.FORMAT: functools.partial(LinearParamsPreparer, '%s'),

    ParamStyle.NUMERIC: NumericParamsPreparer,

    ParamStyle.NAMED: functools.partial(NamedParamsPreparer, NamedParamsPreparer.render_named),
    ParamStyle.PYFORMAT: functools.partial(NamedParamsPreparer, NamedParamsPreparer.render_pyformat),
}


def make_params_preparer(style: ParamStyle) -> ParamsPreparer:
    try:
        fac = _PARAMS_PREPARER_FACTORIES_BY_STYLE[style]
    except KeyError:
        raise ValueError(style) from None
    else:
        return fac()


##


class UnconsumedParamsError(Exception):
    pass


def check_params_consumed(expected: ta.Iterable[K], actual: ta.Iterable[K]) -> None:
    l = set(expected)
    r = set(actual)
    d = l ^ r
    if not d:
        return
    raise UnconsumedParamsError(d)


##


@ta.overload
def substitute_params(
        params: ta.Sequence[K],
        values: ta.Mapping[K, V],
        *,
        strict: bool = False,
) -> ta.Sequence[V]:
    ...


@ta.overload
def substitute_params(
        params: ta.Mapping[str, K],
        values: ta.Mapping[K, V],
        *,
        strict: bool = False,
) -> ta.Mapping[str, V]:
    ...


def substitute_params(params, values, *, strict=False):
    if isinstance(params, collections.abc.Mapping):
        if strict:
            check_params_consumed(values, params.values())

        return {k: values[v] for k, v in params.items()}

    elif isinstance(params, collections.abc.Sequence):
        if strict:
            check_params_consumed(values, params)

        return [values[a] for a in params]

    else:
        raise TypeError(params)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ....lite.abstract import Abstract
from ....lite.namespaces import NamespaceClass
from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineHandlerFn


F = ta.TypeVar('F')
T = ta.TypeVar('T')


##


class IoPipelineHandlerFns(NamespaceClass):
    @dc.dataclass(frozen=True)
    class NoContext(ta.Generic[F, T]):
        fn: ta.Callable[[F], T]

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.fn!r})'

        def __call__(self, ctx: IoPipelineHandlerContext, obj: F) -> T:
            return self.fn(obj)

    @classmethod
    def no_context(cls, fn: ta.Callable[[F], T]) -> IoPipelineHandlerFn[F, T]:
        return cls.NoContext(fn)

    #

    @dc.dataclass(frozen=True)
    class And:
        fns: ta.Sequence[IoPipelineHandlerFn[ta.Any, bool]]

        def __repr__(self) -> str:
            return f'{type(self).__name__}([{", ".join(map(repr, self.fns))}])'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
            return all(fn(ctx, msg) for fn in self.fns)

    @classmethod
    def and_(cls, *fns: IoPipelineHandlerFn[ta.Any, bool]) -> IoPipelineHandlerFn[ta.Any, bool]:
        if len(fns) == 1:
            return fns[0]
        return cls.And(fns)

    #

    @dc.dataclass(frozen=True)
    class Or:
        fns: ta.Sequence[IoPipelineHandlerFn[ta.Any, bool]]

        def __repr__(self) -> str:
            return f'{type(self).__name__}([{", ".join(map(repr, self.fns))}])'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
            return any(fn(ctx, msg) for fn in self.fns)

    @classmethod
    def or_(cls, *fns: IoPipelineHandlerFn[ta.Any, bool]) -> IoPipelineHandlerFn[ta.Any, bool]:
        if len(fns) == 1:
            return fns[0]
        return cls.Or(fns)

    #

    @dc.dataclass(frozen=True)
    class Not:
        fn: IoPipelineHandlerFn[ta.Any, bool]

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.fn!r})'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
            return not self.fn(ctx, msg)

    @classmethod
    def not_(cls, fn: IoPipelineHandlerFn[ta.Any, bool]) -> IoPipelineHandlerFn[ta.Any, bool]:
        if isinstance(fn, cls.Not):
            return fn.fn
        return cls.Not(fn)

    #

    @dc.dataclass(frozen=True)
    class IsInstance:
        ty: ta.Union[type, ta.Tuple[type, ...]]

        def __repr__(self) -> str:
            return f'{type(self).__name__}({self.ty!r})'

        def __call__(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
            return isinstance(msg, self.ty)

    @classmethod
    def isinstance(cls, ty: ta.Union[type, ta.Tuple[type, ...]]) -> IoPipelineHandlerFn[ta.Any, bool]:
        return cls.IsInstance(ty)

    @classmethod
    def not_isinstance(cls, ty: ta.Union[type, ta.Tuple[type, ...]]) -> IoPipelineHandlerFn[ta.Any, bool]:
        return cls.Not(cls.IsInstance(ty))


##


class FnIoPipelineHandler(IoPipelineHandler, Abstract):
    @classmethod
    def of(
            cls,
            *,
            inbound: ta.Optional[IoPipelineHandlerFn[ta.Any, None]] = None,
            outbound: ta.Optional[IoPipelineHandlerFn[ta.Any, None]] = None,
    ) -> IoPipelineHandler:
        if inbound is not None and outbound is not None:
            return DuplexFnIoPipelineHandler(inbound=inbound, outbound=outbound)
        elif inbound is not None:
            return InboundFnIoPipelineHandler(inbound)
        elif outbound is not None:
            return OutboundFnIoPipelineHandler(outbound)
        else:
            raise ValueError('At least one of inbound or outbound must be specified')


class InboundFnIoPipelineHandler(FnIoPipelineHandler):
    def __init__(self, inbound: IoPipelineHandlerFn[ta.Any, None]) -> None:
        super().__init__()

        self._inbound = inbound

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}({self._inbound!r})'

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._inbound(ctx, msg)


class OutboundFnIoPipelineHandler(FnIoPipelineHandler):
    def __init__(self, outbound: IoPipelineHandlerFn[ta.Any, None]) -> None:
        super().__init__()

        self._outbound = outbound

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}({self._outbound!r})'

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._outbound(ctx, msg)


class DuplexFnIoPipelineHandler(FnIoPipelineHandler):
    def __init__(
            self,
            *,
            inbound: IoPipelineHandlerFn[ta.Any, None],
            outbound: IoPipelineHandlerFn[ta.Any, None],
    ) -> None:
        super().__init__()

        self._inbound = inbound
        self._outbound = outbound

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}(inbound={self._inbound!r}, outbound={self._outbound!r})'

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._inbound(ctx, msg)

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._outbound(ctx, msg)



import operator
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .keywords import CanKeyword
from .keywords import Keyword
from .keywords import Star
from .names import CanName
from .names import Name


##


FuncArg: ta.TypeAlias = ta.Union[  # noqa
    Expr,
    Star,
]


class Func(Expr, lang.Final):
    name: Keyword | Name
    args: ta.Sequence[FuncArg] = dc.xfield((), coerce=tuple, repr_fn=lang.truthy_repr) | msh.with_field_options(omit_if=operator.not_)  # noqa


CanFuncArg: ta.TypeAlias = FuncArg | CanExpr


class FuncArgsAccessor:
    def __init__(self, fb: 'FuncBuilder', k: CanKeyword) -> None:
        self.__fb = fb
        self.__k = k

    def __call__(self, *args: CanExpr) -> Func:
        return self.__fb.func(self.__fb.keyword(self.__k), *args)


class FuncAccessor:
    def __init__(self, fb: 'FuncBuilder') -> None:
        self.__fb = fb

    def __getattr__(self, k: CanKeyword) -> FuncArgsAccessor:
        return FuncArgsAccessor(self.__fb, k)


class FuncBuilder(ExprBuilder):
    def func_arg(self, o: CanFuncArg) -> FuncArg:
        return o if isinstance(o, Star) else self.expr(o)

    def func(
            self,
            func: Keyword | CanName,
            *args: CanFuncArg,
    ) -> Func:
        return Func(
            func if isinstance(func, Keyword) else self.name(func),
            tuple(map(self.func_arg, args)),
        )

    @property
    def f(self) -> FuncAccessor:
        return FuncAccessor(self)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
https://man7.org/linux/man-pages/man8/lsof.8.html
"""
import dataclasses as dc
import enum
import typing as ta

from ...lite.check import check
from ...lite.dataclasses import dataclass_repr_omit_falsey
from ...lite.marshal import OBJ_MARSHALER_OMIT_IF_NONE
from ...subprocesses.run import SubprocessRun
from ...subprocesses.run import SubprocessRunnable
from ...subprocesses.run import SubprocessRunOutput


##


# FIXME: ??
# https://unix.stackexchange.com/a/86011
class LsofItemMode(enum.Enum):
    SOLARIS_NFS_LOCK = 'N'
    READ_LOCK_PARTIAL = 'r'
    READ_LOCK_FULL = 'R'
    WRITE_LOCK_PARTIAL = 'w'
    WRITE_LOCK_FULL = 'W'
    READ_WRITE_LOCK = 'u'
    UNKNOWN_LOCK_TYPE = 'U'
    XENIX_LOCK_PARTIAL = 'x'
    XENIX_LOCK_FULL = 'X'


@dc.dataclass(frozen=True)
class _LsofFieldMeta:
    prefix: str
    process: bool = False
    variadic: bool = False

    @classmethod
    def make(cls, *args, **kwargs):
        return {
            cls: cls(*args, **kwargs),
            OBJ_MARSHALER_OMIT_IF_NONE: True,
        }


@dc.dataclass(frozen=True)
class LsofItem:
    __repr__ = dataclass_repr_omit_falsey

    class _PREFIX:
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    class _PROCESS:
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    pid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('p', process=True))
    pgid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('g', process=True))
    ppid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('R', process=True))
    cmd: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('c', process=True))

    uid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('u', process=True))
    login_name: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('L', process=True))

    # FD is the File Descriptor number of the file or:
    #   cwd  current working directory;
    #   Lnn  library references (AIX);
    #   ctty character tty;
    #   DEL  deleted file;
    #   err  FD information error (see NAME column);
    #   fp.  Fileport (Darwin);
    #   jld  jail directory (FreeBSD);
    #   ltx  shared library text (code and data);
    #   Mxx  hex memory-mapped type number xx.
    #   m86  DOS Merge mapped file;
    #   mem  memory-mapped file;
    #   mmap memory-mapped device;
    #   NOFD for a Linux /proc/<PID>/fd directory that can't be opened -- the directory path appears in the NAME column,
    #     followed by an error message;
    #   pd   parent directory;
    #   Rnn  unknown pregion number (HP-UX);
    #   rtd  root directory;
    #   twd  per task current working directory;
    #   txt  program text (code and data);
    #   v86  VP/ix mapped file;
    #
    # FD is followed by one of these characters, describing the mode under which the file is open:
    #   r for read access;
    #   w for write access;
    #   u for read and write access;
    #   space if mode unknown and no lock
    #   character follows;
    #   `-' if mode unknown and lock
    #   character follows.
    #
    # The mode character is followed by one of these lock characters, describing the type of lock applied to the file:
    #   N for a Solaris NFS lock of unknown type;
    #   r for read lock on part of the file;
    #   R for a read lock on the entire file;
    #   w for a write lock on part of the file;
    #   W for a write lock on the entire file;
    #   u for a read and write lock of any length;
    #   U for a lock of unknown type;
    #   x for an SCO OpenServer Xenix lock on part of the file;
    #   X for an SCO OpenServer Xenix lock on the entire file;
    #   space if there is no lock.
    #
    fd: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('f'))

    inode: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('i'))
    name: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('n'))

    # r = read; w = write; u = read/write
    access: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('a'))

    # r/R = read; w/W = write; u = read/write
    lock: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('l'))

    file_flags: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('G'))
    file_type: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('t'))
    file_size: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('s'))
    file_offset: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('o'))  # as 0t<dec> or 0x<hex>

    device_character_code: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('d'))
    device_number: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('D'))  # as 0x<hex>
    raw_device_number: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('r'))  # as 0x<hex>

    share_count: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('C'))
    link_count: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('k'))

    stream_info: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('S'))
    protocol_name: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('P'))

    tcp_tpi_info: ta.Optional[ta.Sequence[str]] = dc.field(
        default=None,
        metadata=_LsofFieldMeta.make('T', variadic=True),
    )

    # 'K':  'task_id',  # unknown field
    # 'M':  'task_command_name',  # unknown field

    _FIELDS_BY_PREFIX: ta.ClassVar[ta.Mapping[str, dc.Field]]
    _DEFAULT_PREFIXES: ta.ClassVar[str]

    @classmethod
    def from_prefixes(cls, dct: ta.Mapping[str, ta.Any]) -> 'LsofItem':
        kw: ta.Dict[str, ta.Any] = {
            fld.name: val
            for pfx, val in dct.items()
            if (fld := cls._FIELDS_BY_PREFIX.get(pfx)) is not None
        }

        return cls(**kw)

    @classmethod
    def from_prefix_lines(cls, lines: ta.Iterable[str]) -> ta.List['LsofItem']:
        proc_dct: ta.Dict[str, ta.Any] = {}
        file_dct: ta.Dict[str, ta.Any] = {}
        out: ta.List[LsofItem] = []

        def emit() -> None:
            if file_dct:
                out.append(cls.from_prefixes({**proc_dct, **file_dct}))

        for line in lines:
            pfx, val = line[0], line[1:]
            try:
                fld = cls._FIELDS_BY_PREFIX[pfx]
            except KeyError:
                continue
            meta: _LsofFieldMeta = fld.metadata[_LsofFieldMeta]

            if pfx == 'p':
                emit()
                proc_dct = {}
                file_dct = {}
            elif pfx == 'f':
                emit()
                file_dct = {}

            if meta.process:
                dct = proc_dct
            else:
                dct = file_dct

            if meta.variadic:
                dct.setdefault(pfx, []).append(val)
            else:
                if pfx in dct:
                    raise KeyError(pfx)
                dct[pfx] = val

        emit()

        return out


LsofItem._FIELDS_BY_PREFIX = {  # noqa
    meta.prefix: fld
    for fld in dc.fields(LsofItem)  # noqa
    if (meta := fld.metadata.get(_LsofFieldMeta)) is not None  # noqa
}

LsofItem._DEFAULT_PREFIXES = ''.join(LsofItem._FIELDS_BY_PREFIX)  # noqa


##


@dc.dataclass(frozen=True)
class LsofCommand(SubprocessRunnable[ta.List[LsofItem]]):
    pid: ta.Optional[int] = None
    file: ta.Optional[str] = None

    prefixes: ta.Optional[ta.Sequence[str]] = None

    def make_run(self) -> SubprocessRun:
        if (prefixes := self.prefixes) is None:
            prefixes = LsofItem._DEFAULT_PREFIXES  # noqa

        return SubprocessRun.of(
            'lsof',
            '-F', ''.join(prefixes),
            *(['-p', str(self.pid)] if self.pid is not None else []),
            *([self.file] if self.file is not None else []),

            stdout='pipe',
            stderr='devnull',
            check=True,
        )

    def handle_run_output(self, output: SubprocessRunOutput) -> ta.List[LsofItem]:
        lines = [s for l in check.not_none(output.stdout).decode().splitlines() if (s := l.strip())]
        return LsofItem.from_prefix_lines(lines)


##


if __name__ == '__main__':
    def _main() -> None:
        import argparse  # noqa
        import importlib  # noqa
        import json  # noqa

        parser = argparse.ArgumentParser()
        parser.add_argument('--pid', '-p', type=int)
        parser.add_argument('file', nargs='?')
        args = parser.parse_args()

        subprocesses = importlib.import_module('..subprocesses.sync', package=__package__).subprocesses
        items = LsofCommand(
            pid=args.pid,
            file=args.file,
        ).run(subprocesses)

        marshal_obj = importlib.import_module('..lite.marshal', package=__package__).marshal_obj
        print(json.dumps(marshal_obj(items), indent=2))

    _main()



import abc

from ...... import lang


##


class AsyncsBackend(lang.Abstract):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def is_available(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def is_imported(self) -> bool:
        raise NotImplementedError

    #

    def prepare_for_metafunc(self, metafunc) -> None:  # noqa
        pass

    #

    @abc.abstractmethod
    def wrap_runner(self, fn):
        raise NotImplementedError

    async def install_context(self, contextvars_ctx):  # noqa
        pass



# ruff: noqa: I001
from .. import dataclasses as _dc  # noqa


_dc.init_package(
    globals(),
    codegen=True,
)


##


from .. import lang as _lang  # noqa


with _lang.auto_proxy_init(globals()):
    ##

    from .api.configs import (  # noqa
        Config,
        Configs,

        ConfigRegistry,
    )

    from .api.contexts import (  # noqa
        BaseContext,

        MarshalFactoryContext,
        UnmarshalFactoryContext,

        MarshalContext,
        UnmarshalContext,
    )

    from .api.errors import (  # noqa
        ForbiddenTypeError,
        MarshalError,
        UnhandledTypeError,
    )

    from .api.funcs import (  # noqa
        FuncMarshaler,
        FuncUnmarshaler,

        FuncMarshalerFactory,
        FuncUnmarshalerFactory,
    )

    from .api.naming import (  # noqa
        Naming,
        translate_name,
    )

    from .api.options import (  # noqa
        Option,
    )

    from .api.overrides import (  # noqa
        Override,
        ReflectOverride,
    )

    from .api.registries import (  # noqa
        RegistrySealedError,
        Registry,
    )

    from .api.types import (  # noqa
        Marshaler,
        Unmarshaler,

        MarshalerMaker,
        UnmarshalerMaker,

        MarshalerFactory,
        UnmarshalerFactory,

        Marshaling,
    )

    from .api.values import (  # noqa
        Value,
    )

    from .composite.iterables import (  # noqa
        IterableMarshaler,
        IterableUnmarshaler,
    )

    from .composite.optionals import (  # noqa
        OptionalMarshaler,
        OptionalUnmarshaler,
    )

    from .composite.unions.literals import (  # noqa
        LITERAL_UNION_TYPES,
        LiteralUnionMarshaler,
        LiteralUnionMarshalerFactory,
        LiteralUnionUnmarshaler,
        LiteralUnionUnmarshalerFactory,
    )

    from .composite.unions.primitives import (  # noqa
        PRIMITIVE_UNION_TYPES,
        PrimitiveUnionMarshaler,
        PrimitiveUnionMarshalerFactory,
        PrimitiveUnionUnmarshaler,
        PrimitiveUnionUnmarshalerFactory,
    )

    from .composite.wrapped import (  # noqa
        WrappedMarshaler,
        WrappedUnmarshaler,
    )

    from .factories.method import (  # noqa
        MarshalerFactoryMethodClass,
        UnmarshalerFactoryMethodClass,
    )

    from .factories.moduleimport.api import (  # noqa
        ModuleImport,
    )

    from .factories.moduleimport.factories import (  # noqa
        ModuleImportingMarshalerFactory,
        ModuleImportingUnmarshalerFactory,
    )

    from .factories.multi import (  # noqa
        MultiMarshalerFactory,
        MultiUnmarshalerFactory,
    )

    from .factories.typemap import (  # noqa
        TypeMapMarshalerFactory,
        TypeMapUnmarshalerFactory,
    )

    from .factories.typecache import (  # noqa
        TypeCacheMarshalerFactory,
        TypeCacheUnmarshalerFactory,
    )

    from .factories.recursive import (  # noqa
        RecursiveMarshalerFactory,
        RecursiveUnmarshalerFactory,
    )

    from .objects.dataclasses import (  # noqa
        AbstractDataclassFactory,
        DataclassMarshalerFactory,
        DataclassUnmarshalerFactory,
        get_dataclass_field_infos,
        get_dataclass_options,
    )

    from .objects.api import (  # noqa
        FieldOptions,
        ObjectOptions,
        ObjectSpecials,
    )

    from .objects.helpers import (  # noqa
        update_fields_options,
        update_object_options,
        with_field_options,
    )

    from .objects.infos import (  # noqa
        FieldInfo,
        FieldInfos,
    )

    from .objects.marshal import (  # noqa
        ObjectMarshaler,
        SimpleObjectMarshalerFactory,
    )

    from .objects.unmarshal import (  # noqa
        ObjectUnmarshaler,
        SimpleObjectUnmarshalerFactory,
    )

    from .polymorphism.api import (  # noqa
        AutoStripSuffix,
        FieldTypeTagging,
        Impl,
        Impls,
        ImplBase,
        ImplBases,
        Polymorphism,
        TypeTagging,
        WrapperTypeTagging,
        polymorphism_from_impls,
        polymorphism_from_subclasses,
    )

    from .polymorphism.marshal import (  # noqa
        PolymorphismMarshaler,
        PolymorphismMarshalerFactory,
        make_polymorphism_marshaler,
    )

    from .polymorphism.standard import (  # noqa
        standard_polymorphism_factories,
    )

    from .polymorphism.unmarshal import (  # noqa
        PolymorphismUnmarshaler,
        PolymorphismUnmarshalerFactory,
        make_polymorphism_unmarshaler,
    )

    from .singular.base64 import (  # noqa
        BASE64_MARSHALER_FACTORY,
        BASE64_UNMARSHALER_FACTORY,
        Base64MarshalerUnmarshaler,
    )

    from .singular.primitives import (  # noqa
        PRIMITIVE_TYPES,
    )

    from .trivial.forbidden import (  # noqa
        ForbiddenTypeMarshalerFactory,
        ForbiddenTypeMarshalerFactoryUnmarshalerFactory,
        ForbiddenTypeUnmarshalerFactory,
    )

    from .trivial.nop import (  # noqa
        NOP_MARSHALER_UNMARSHALER,
        NopMarshalerUnmarshaler,
    )

    from .globals import (  # noqa
        global_config_registry,
        global_marshaler_factory,
        global_unmarshaler_factory,
        global_marshaling,

        marshal,
        unmarshal,

        register_global_config,
        register_global_module_import,
    )

    from .standard import (  # noqa
        StandardFactories,
        DEFAULT_STANDARD_FACTORIES,

        new_standard_marshaler_factory,
        new_standard_unmarshaler_factory,

        install_standard_factories,
    )



from .managers import (  # noqa
    ResourceManagerRef,
    ResourceManagerRefNotRegisteredError,

    BaseResourceManager,
    BaseResourceManaged,

    ResourceManager,
    ResourceManaged,

    AsyncResourceManager,
    AsyncResourceManaged,
)

from .debug import (  # noqa
    get_resource_debug,
    set_resource_debug,
)

from .errors import (  # noqa
    ResourceNotEnteredError,
    UnclosedResourceWarning,
)

from .simple import (  # noqa
    BaseSimpleResource,
    SimpleResource,
    AsyncSimpleResource,
)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
Notes:
 - necessarily string-oriented
 - single file, as this is intended to be amalg'd and thus all included anyway

TODO:
 - ConfigDataMapper? to_map -> ConfigMap?
 - nginx ?
 - raw ?
"""
import abc
import collections.abc
import configparser
import dataclasses as dc
import json
import os.path
import typing as ta

from ..formats.ini.sections import IniSectionSettingsMap
from ..formats.ini.sections import extract_ini_sections
from ..formats.ini.sections import render_ini_sections
from ..formats.toml.parser import toml_loads
from ..formats.toml.writer import TomlWriter
from ..formats.yaml.backends import DEFAULT_YAML_BACKEND
from ..formats.yaml.backends import YamlBackend
from ..lite.abstract import Abstract
from ..lite.check import check
from ..lite.json import json_dumps_pretty
from .types import ConfigMap


T = ta.TypeVar('T')

ConfigDataT = ta.TypeVar('ConfigDataT', bound='ConfigData')
ObjConfigDataT = ta.TypeVar('ObjConfigDataT', bound='ObjConfigData')


##


@dc.dataclass(frozen=True)
class ConfigData(Abstract):
    @abc.abstractmethod
    def as_map(self) -> ConfigMap:
        raise NotImplementedError


#


class ConfigLoaderContext:
    pass


class ConfigFileLoader(Abstract, ta.Generic[ConfigDataT]):
    @property
    @abc.abstractmethod
    def data_cls(self) -> ta.Type[ConfigDataT]:
        raise NotImplementedError

    @property
    def file_exts(self) -> ta.Sequence[str]:
        return ()

    def match_file(self, n: str) -> bool:
        return '.' in n and n.split('.', maxsplit=1)[-1] in check.not_isinstance(self.file_exts, str)

    def find_file(self, p: str) -> ta.Optional[str]:
        hits: ta.List[str] = []
        for e in self.file_exts:
            cur = f'{p}.{e}'
            if os.path.exists(cur):
                check.state(os.path.isfile(cur))
                hits.append(cur)
        if hits:
            return check.single(hits)
        return None

    @abc.abstractmethod
    def load_file(self, p: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ConfigDataT:
        raise NotImplementedError


class ConfigLoader(ConfigFileLoader[ConfigDataT], Abstract, ta.Generic[ConfigDataT]):
    def load_file(self, p: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ConfigDataT:
        with open(p) as f:
            return self.load_str(f.read(), ctx)

    @abc.abstractmethod
    def load_str(self, s: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ConfigDataT:
        raise NotImplementedError


#


class ProxyConfigFileLoader(ConfigFileLoader[ConfigDataT]):
    def __init__(self, underlying: ta.Union[ConfigLoader[ConfigDataT], ta.Callable[[], ConfigLoader[ConfigDataT]]]) -> None:  # noqa
        super().__init__()

        self._underlying = underlying

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._underlying!r})'

    @property
    def underlying(self) -> ta.Union[ConfigLoader[ConfigDataT], ta.Callable[[], ConfigLoader[ConfigDataT]]]:
        return self._underlying

    def get_underlying(self) -> ConfigLoader[ConfigDataT]:
        if callable(ul := self._underlying):
            ul = ul()
        return ul

    #

    @property
    def data_cls(self) -> ta.Type[ConfigDataT]:
        return self.get_underlying().data_cls

    @property
    def file_exts(self) -> ta.Sequence[str]:
        return self.get_underlying().file_exts

    def match_file(self, n: str) -> bool:
        return self.get_underlying().match_file(n)

    def find_file(self, p: str) -> ta.Optional[str]:
        return self.get_underlying().find_file(p)

    def load_file(self, p: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ConfigDataT:
        return self.get_underlying().load_file(p, ctx)


class ProxyConfigLoader(ProxyConfigFileLoader[ConfigDataT], ConfigLoader[ConfigDataT]):
    def load_str(self, s: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ConfigDataT:
        return self.get_underlying().load_str(s, ctx)


#


class ConfigRendererContext:
    pass


class ConfigRenderer(Abstract, ta.Generic[ConfigDataT]):
    @property
    @abc.abstractmethod
    def data_cls(self) -> ta.Type[ConfigDataT]:
        raise NotImplementedError

    def match_data(self, d: ConfigDataT) -> bool:
        return isinstance(d, self.data_cls)

    #

    @abc.abstractmethod
    def render(self, d: ConfigDataT, ctx: ta.Optional[ConfigRendererContext] = None) -> str:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class ObjConfigData(ConfigData, Abstract):
    obj: ta.Any

    def as_map(self) -> ConfigMap:
        return check.isinstance(self.obj, collections.abc.Mapping)


##


class LoadsConfigLoader(ConfigLoader[ObjConfigDataT], Abstract, ta.Generic[ObjConfigDataT]):
    @abc.abstractmethod
    def loads(self, s: str) -> ta.Any:
        raise NotImplementedError

    def load_str(self, s: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ObjConfigDataT:
        return self.data_cls(self.loads(s))


class DumpsConfigRenderer(ConfigRenderer[ObjConfigDataT], Abstract, ta.Generic[ObjConfigDataT]):
    @abc.abstractmethod
    def dumps(self, o: ta.Any) -> str:
        raise NotImplementedError

    def render(self, d: ObjConfigDataT, ctx: ta.Optional[ConfigRendererContext] = None) -> str:
        return self.dumps(d.obj)


##


@dc.dataclass(frozen=True)
class JsonConfigData(ObjConfigData):
    pass


class JsonConfigLoader(LoadsConfigLoader[JsonConfigData]):
    data_cls = JsonConfigData
    file_exts = ('json',)

    def loads(self, s: str) -> ta.Any:
        return json.loads(s)


class JsonConfigRenderer(DumpsConfigRenderer[JsonConfigData]):
    data_cls = JsonConfigData

    def dumps(self, o: ta.Any) -> str:
        return json_dumps_pretty(o)


##


@dc.dataclass(frozen=True)
class TomlConfigData(ObjConfigData):
    pass


class TomlConfigLoader(LoadsConfigLoader[TomlConfigData]):
    data_cls = TomlConfigData
    file_exts = ('toml',)

    def loads(self, s: str) -> ta.Any:
        return toml_loads(s)


class TomlConfigRenderer(DumpsConfigRenderer[TomlConfigData]):
    data_cls = TomlConfigData

    def dumps(self, o: ta.Any) -> str:
        return TomlWriter.write_str(o)


##


@dc.dataclass(frozen=True)
class YamlConfigData(ObjConfigData):
    pass


class YamlConfigLoader(LoadsConfigLoader[YamlConfigData]):
    data_cls = YamlConfigData
    file_exts = ('yaml', 'yml')
    backend: ta.Optional[YamlBackend] = None

    def loads(self, s: str) -> ta.Any:
        return (self.backend or DEFAULT_YAML_BACKEND.INSTANCE).loads(s)


class YamlConfigRenderer(DumpsConfigRenderer[YamlConfigData]):
    data_cls = YamlConfigData
    backend: ta.Optional[YamlBackend] = None

    def dumps(self, o: ta.Any) -> str:
        return (self.backend or DEFAULT_YAML_BACKEND.INSTANCE).dumps(o)


##


@dc.dataclass(frozen=True)
class IniConfigData(ConfigData):
    sections: IniSectionSettingsMap

    def as_map(self) -> ConfigMap:
        return self.sections


class IniConfigLoader(ConfigLoader[IniConfigData]):
    data_cls = IniConfigData
    file_exts = ('ini',)

    def load_str(self, s: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> IniConfigData:
        cp = configparser.ConfigParser()
        cp.read_string(s)
        return IniConfigData(extract_ini_sections(cp))


class IniConfigRenderer(ConfigRenderer[IniConfigData]):
    data_cls = IniConfigData

    def render(self, d: IniConfigData, ctx: ta.Optional[ConfigRendererContext] = None) -> str:
        return render_ini_sections(d.sections)


##


DEFAULT_CONFIG_LOADERS: ta.Sequence[ConfigLoader] = [
    JsonConfigLoader(),
    TomlConfigLoader(),
    YamlConfigLoader(),
    IniConfigLoader(),
]

DEFAULT_CONFIG_LOADER: ConfigLoader = JsonConfigLoader()


#


@dc.dataclass(frozen=True)
class SwitchedConfigFileLoader(ConfigFileLoader[ConfigData]):
    loaders: ta.Sequence[ConfigLoader]
    default: ta.Optional[ConfigLoader] = None

    def __post_init__(self) -> None:
        seen: ta.Set[str] = set()
        for l in self.loaders:
            for e in l.file_exts:
                if e in seen:
                    raise ValueError(e)
                seen.add(e)

    @property
    def data_cls(self) -> ta.Type[ConfigData]:
        return ConfigData

    @property
    def file_exts(self) -> ta.Sequence[str]:
        return [e for l in self.loaders for e in l.file_exts]

    def load_file(self, p: str, ctx: ta.Optional[ConfigLoaderContext] = None) -> ConfigData:
        n = os.path.basename(p)

        for l in self.loaders:
            if l.match_file(n):
                return l.load_file(p, ctx)

        if (d := self.default) is not None:
            return d.load_file(p, ctx)

        raise NameError(n)


DEFAULT_CONFIG_FILE_LOADER = SwitchedConfigFileLoader(
    loaders=DEFAULT_CONFIG_LOADERS,
    default=DEFAULT_CONFIG_LOADER,
)


##


@dc.dataclass(frozen=True)
class SwitchedConfigRenderer(ConfigRenderer[ConfigData]):
    renderers: ta.Sequence[ConfigRenderer]

    @property
    def data_cls(self) -> ta.Type[ConfigData]:
        return ConfigData

    def render(self, d: ConfigData, ctx: ta.Optional[ConfigRendererContext] = None) -> str:
        for r in self.renderers:
            if r.match_data(d):
                return r.render(d, ctx)
        raise TypeError(d)


DEFAULT_CONFIG_RENDERERS: ta.Sequence[ConfigRenderer] = [
    JsonConfigRenderer(),
    TomlConfigRenderer(),
    YamlConfigRenderer(),
    IniConfigRenderer(),
]

DEFAULT_CONFIG_RENDERER = SwitchedConfigRenderer(DEFAULT_CONFIG_RENDERERS)



import typing as ta

from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from .base import Node
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .idents import Ident
from .keywords import Star
from .relations import CanRelation
from .relations import Relation
from .relations import RelationBuilder
from .stmts import Stmt


##


class SelectItem(Node, lang.Abstract):
    pass


class AllSelectItem(SelectItem, lang.Final):
    pass


class ExprSelectItem(SelectItem, lang.Final):
    v: Expr
    a: Ident | None = dc.xfield(None, repr_fn=lang.opt_repr) | msh.with_field_options(omit_if=lang.is_none)


##


class Select(Stmt, lang.Final):
    items: ta.Sequence[SelectItem] = dc.xfield(coerce=tuple)
    from_: Relation | None = dc.xfield(None, repr_fn=lang.opt_repr) | msh.with_field_options(name='from', omit_if=lang.is_none)  # noqa
    where: Expr | None = dc.xfield(None, repr_fn=lang.opt_repr) | msh.with_field_options(omit_if=lang.is_none)


class SelectExpr(Expr, lang.Final):
    s: Select


class SelectRelation(Relation, lang.Final):
    s: Select


##


CanSelectItem: ta.TypeAlias = SelectItem | Star | CanExpr


class SelectBuilder(RelationBuilder, ExprBuilder):
    def select_item(self, o: CanSelectItem) -> SelectItem:
        if isinstance(o, SelectItem):
            return o
        elif isinstance(o, Star):
            return AllSelectItem()
        else:
            return ExprSelectItem(self.expr(o))

    def select(
            self,
            items: ta.Sequence[CanSelectItem],
            from_: CanRelation | None = None,
            where: CanExpr | None = None,
    ) -> Select:
        return Select(
            tuple(self.select_item(i) for i in items),
            from_=self.relation(from_) if from_ is not None else None,
            where=self.expr(where) if where is not None else None,
        )



import dataclasses as dc
import termios


##


@dc.dataclass()
class TermState:
    iflag: int
    oflag: int
    cflag: int
    lflag: int
    ispeed: int
    ospeed: int
    cc: list[bytes]

    #

    def copy(self) -> 'TermState':
        return TermState(*self.as_list())  # type: ignore[arg-type]

    def as_list(self) -> list[int | list[bytes]]:
        return [
            self.iflag,
            self.oflag,
            self.cflag,
            self.lflag,
            self.ispeed,
            self.ospeed,
            self.cc[:],
        ]


def get_term_state(fd: int) -> TermState:
    return TermState(*termios.tcgetattr(fd))


def set_term_state(fd: int, attrs: TermState, when: int = termios.TCSANOW) -> None:
    termios.tcsetattr(fd, when, attrs.as_list())


##


class TermStateStack:
    def __init__(self, fd: int = 0) -> None:  # noqa
        super().__init__()

        self._fd = fd
        self._stack: list[TermState] = []

    #

    def get(self) -> TermState:
        return get_term_state(self._fd)

    def set(self, attrs: TermState, when: int = termios.TCSANOW) -> None:
        set_term_state(self._fd, attrs, when)

    #

    def save(self) -> None:
        self._stack.append(self.get())

    def restore(self) -> None:
        self.set(self._stack.pop())
